"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing abstract superclasses for validation.

Routine listings
----------------
MatrixBase(object)
    Abstract base class of custom matrix classes.
MMSEInputChannel(object)
    Abstract base class of a Minimum Mean Squaread Error (MMSE) input channel.
StopCriterion(object)
    Abstract base class of a stop criterion.
ThresholdOperator(object)
    Abstract base class of a threshold operator.

"""


class MatrixBase(object):
    """
    Abstract base class of custom matrix classes.

    The `magni.utils.validation.validate_numeric` function accepts built-in
    numeric types, numpy built-in numeric types, and subclasses of the present
    class. In order to perform validation checks, the validation function needs
    to know the data type, the bounds, and the shape of the variable. Thus,
    subclasses must call the init function of the present class with these
    arguments.

    Parameters
    ----------
    dtype : type
        The data type of the values of the instance.
    bounds : list or tuple
        The bounds of the values of the instance.
    shape : list or tuple
        The shape of the instance.

    Attributes
    ----------
    bounds : list or tuple
        The bounds of the values of the instance.
    dtype : type
        The data type of the values of the instance.
    shape : list or tuple
        The shape of the instance.

    Notes
    -----
    `dtype` is either a built-in numeric type or a numpy built-in numeric type.

    If the matrix has complex values, `bounds` is a list with two values; The
    bounds on the real values and the bounds on the imaginary values. If, on
    the other hand, the matrix has real values, `bounds` has one value; The
    bounds on the real values. Each such bounds value is a list with two real,
    numeric values; The lower bound (that is, the minimum value) and the upper
    bound (that is, the maximum value).

    """

    def __init__(self, dtype, bounds, shape):
        self._dtype = dtype
        self._bounds = bounds
        self._shape = shape

    bounds = property(lambda self: self._bounds)
    dtype = property(lambda self: self._dtype)
    shape = property(lambda self: self._shape)


class MMSEInputChannel(object):
    """
    Abstract base class of a Minimum Mean Squaread Error (MMSE) input channel.

    The `magni.cs.reconstruction` algorithms may make use of input channels to
    define the prior knowledge on the sought solution. In order for the
    validation of such an input channel to work, it must be based on this
    class.

    Parameters
    ----------
    var : dict
        The input channel state variables.

    """

    def __init__(self, var):
        pass

    def compute(self, var):
        """
        Compute the input channel value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the input channel value.

        Returns
        -------
        mean : ndarray
            The computed input channel mean.
        variance : ndarray
            The computed input channel variance.

        """

        raise NotImplementedError(
            'Subclasses of MMSEInputChannel must override this method.')


class MMSEOutputChannel(object):
    """
    Abstract base class of a Minimum Mean Squaread Error (MMSE) output channel.

    The `magni.cs.reconstruction` algorithms may make use of output channels to
    define the observation model. In order for the validation of such an output
    channel to work, it must be based on this class.

    Parameters
    ----------
    var : dict
        The output channel state variables.

    """

    def __init__(self, var):
        pass

    def compute(self, var):
        """
        Compute the output channel value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the output channel value.

        Returns
        -------
        mean : ndarray
            The computed output channel mean.
        variance : ndarray
            The computed output channel variance.

        """

        raise NotImplementedError(
            'Subclasses of MMSEOutputChannel must override this method.')


class StopCriterion(object):
    """
    Abstract base class of a stop criterion.

    The `magni.cs.reconstruction` algorithms are typically iterative algorithms
    that make use of some stop criterion to determine if it has converged. In
    order for the validation of such a stop criterion to work, it must be based
    on this class.

    Parameters
    ----------
    var : dict
        The stop criterion state variables.

    """

    def __init__(self, var):
        pass

    def compute(self, var):
        """
        Compute the stop criterion value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the stop criterion value.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        value : float
            The stop criterion value.

        """

        raise NotImplementedError(
            'Subclasses of StopCriterion must override this method.')


class ThresholdOperator(object):
    """
    Abstract base class of a threshold operator.

    The `magni.cs.reconstruction` algorithms may make use of threshold
    operators for "de-noising". In order for the validation of such a threshold
    operator to work, it must be based on this class.

    Parameters
    ----------
    var : dict
        The threshold operator state variables.

    """

    def __init__(self, var):
        pass

    def compute_deriv_threshold(self, var):
        """
        Compute the entrywise derivative threshold.

        Parameters
        ----------
        var : dict
            The variables used in computing the derivative threshold.

        Returns
        -------
        eta_deriv : ndarray
            The computed entrywise derivative threshold.

        """

        raise NotImplementedError(
            'Subclasses of ThresholdOperator must override this method.')

    def compute_threshold(self, var):
        """
        Compute the entrywise threshold.

        Parameters
        ----------
        var : dict
            The variables used in computing the threshold.

        Returns
        -------
        eta : ndarray
            The computed entrywise threshold.

        """

        raise NotImplementedError(
            'Subclasses of ThresholdOperator must override this method.')

    def update_threshold_level(self, var):
        """
        Update the threshold level state.

        Parameters
        ----------
        var : dict
            The variables used in computing the threshold level update.

        """

        raise NotImplementedError(
            'Subclasses of ThresholdOperator must override this method.')



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing support functionality for the dictionaries subpackage.

Routine listings
----------------
get_function_handle(type\_, transform)
    Function to get a function handle to a transform method.
get_transform_names()
    Function to get a tuple of names of the available transforms.

"""

from __future__ import division
import types

from magni.imaging.dictionaries import _matrices
from magni.imaging.dictionaries import _mtx1D
from magni.imaging.dictionaries import _visualisations
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic


def get_function_handle(type_, transform):
    """
    Return a function handle to a given transform method.

    Parameters
    ----------
    type_ : {'matrix', 'transform_matrix', 'visualisation'}
        Identifier of the type of method to return a handle to.
    transform : str
        Identifier of the transform method to return a handle to.

    Returns
    -------
    f_handle : function
        Handle to `transform`.

    Notes
    -----
    The available types are:

    * matrix : 2D transform matrix object.
    * transform_matrix : 1D transform matrix.
    * visualisation : Coefficient visualisation preparation function.

    Examples
    --------
    For example, return a handle to the matrix method for a DCT:

    >>> from magni.imaging.dictionaries.utils import get_function_handle
    >>> get_function_handle('matrix', 'DCT').__name__
    'get_DCT'

    or return a handle to the 1D DCT transform matrix method:

    >>> get_function_handle('transform_matrix', 'DCT').__name__
    'get_DCT_transform_matrix'

    or return a handle to the visualisation method for a DFT:

    >>> get_function_handle('visualisation', 'DFT').__name__
    'visualise_DFT'

    """

    @_decorate_validation
    def validate_input():
        _generic('type_', 'string', value_in=('matrix', 'transform_matrix',
                                              'visualisation'))
        _generic('transform', 'string')

    @_decorate_validation
    def validate_output():
        _generic('f_handle', 'function')

    validate_input()

    modules = {'matrix': _matrices,
               'transform_matrix': _mtx1D,
               'visualisation': _visualisations}

    if type_ == 'matrix':
        f_handle = vars(modules[type_])['get_' + transform]
    elif type_ == 'transform_matrix':
        f_handle = vars(modules[type_])[
            'get_' + transform + '_transform_matrix']
    elif type_ == 'visualisation':
        f_handle = vars(modules[type_])['visualise_' + transform]

    validate_output()

    return f_handle


def get_transform_names():
    """
    Return a tuple of names of the available transforms.

    Returns
    -------
    names : tuple
        The tuple of names of available transforms.

    Examples
    --------
    For example, get transform names and extract 'DCT' and 'DFT'

    >>> from magni.imaging.dictionaries.utils import get_transform_names
    >>> names = get_transform_names()
    >>> tuple(sorted(name for name in names if name in ('DCT', 'DFT')))
    ('DCT', 'DFT')

    and a handle to corresponding visualisation method for the DCT

    >>> from magni.imaging.dictionaries.utils import get_function_handle
    >>> f_handles = tuple(get_function_handle('visualisation', name)
    ... for name in names)
    >>> tuple(f_h.__name__ for f_h in f_handles if 'DCT' in f_h.__name__)
    ('visualise_DCT',)

    """

    name_candidates = _matrices.__dict__.keys()
    names = [candidate[4:] for candidate in name_candidates
             if candidate[:4] == 'get_']

    return tuple(names)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the public function of the magni.cs.phase_transition
subpackage.

"""

from __future__ import division
import os
import re
import types

from magni.cs.phase_transition import _analysis
from magni.cs.phase_transition import _simulation
from magni.utils.multiprocessing import File as _File
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


def determine(algorithm, path, label='default', overwrite=False,
              pre_simulation_hook=None):
    """
    Determine the phase transition of a reconstruction algorithm.

    The phase transition is determined from a number of monte carlo simulations
    on a delta-rho grid for a given problem suite.

    Parameters
    ----------
    algorithm : function
        A function handle to the reconstruction algorithm.
    path : str
        The path of the HDF5 database where the results should be stored.
    label : str
        The label assigned to the phase transition (the default is 'default').
    overwrite : bool
        A flag indicating if an existing phase transition should be overwritten
        if it has the same path and label (the default is False).
    pre_simulation_hook : callable
        A handle to a callable which should be run *just* before the call to
        the reconstruction algorithm (the default is None, which implies that
        no pre hook is run).

    See Also
    --------
    magni.cs.phase_transition._simulation.run : The actual simulation.
    magni.cs.phase_transition._analysis.run : The actual phase determination.

    Notes
    -----
    The `pre_simulation_hook` may be used to setup the simulation to match the
    specfic simulation parameters, e.g. if an oracle estimator is used in the
    reconstruction algorithm. The `pre_simulation_hook` takes one argument
    which is the locals() dict.

    Examples
    --------
    An example of how to use this function is provided in the `examples` folder
    in the `cs-phase_transition.ipynb` ipython notebook file.

    """

    @_decorate_validation
    def validate_input():
        _generic('algorithm', 'function')
        _generic('path', 'string')
        _generic('label', 'string')

        # regular expression matching invalid characters
        match = re.search(r'[^a-zA-Z0-9 ,.\-_/]', label)

        if match is not None:
            msg = 'The value of >>label<<, {!r}, may not contain {!r}.'
            raise RuntimeError(msg.format(label, match.group()))

        # regular expression matching labels without empty path components
        match = re.search(r'^([^/]+/)*[^/]+$', label)

        if match is None:
            msg = "The value of >>label<<, {!r}, may not contain '' folders."
            raise RuntimeError(msg.format(label))

        _numeric('overwrite', 'boolean')
        if (pre_simulation_hook is not None and
                not callable(pre_simulation_hook)):
            raise RuntimeError('The >>pre_simulation_hook<< must be callable')

    validate_input()

    if os.path.isfile(path):
        with _File(path, 'a') as f:
            if '/' + label in f:
                if overwrite:
                    f.remove_node('/' + label, recursive=True)
                else:
                    raise IOError("{!r} already uses the label {!r}."
                                  .format(path, label))

    _simulation.run(algorithm, path, label,
                    pre_simulation_hook=pre_simulation_hook)
    _analysis.run(path, label)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
square_spiral_sample_image(h, w, scan_length, num_points)
    Function for square spiral sampling an image.
square_spiral_sample_surface(l, w, speed, sample_rate, time)
    Function for square spiral sampling a surface.

"""

from __future__ import division

import numpy as np

from magni.imaging.measurements import _util
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['square_spiral_sample_image', 'square_spiral_sample_surface']

_min_l = _util.min_l
_min_w = _util.min_w
_min_speed = _util.min_speed
_min_sample_rate = _util.min_sample_rate
_min_time = _util.min_time
_min_scan_length = _util.min_scan_length
_min_num_points = _util.min_num_points


def square_spiral_sample_image(h, w, scan_length, num_points):
    """
    Sample an image using a square spiral pattern.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a square spiral pattern are determined. The
    `scan_length` determines the length of the path scanned whereas
    `num_points` indicates the number of samples taken on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import square_spiral_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> np.set_printoptions(suppress=True)
    >>> square_spiral_sample_image(h, w, scan_length, num_points)
    array([[ 5.        ,  5.        ],
           [ 6.28571429,  5.97619048],
           [ 4.38095238,  3.71428571],
           [ 2.42857143,  5.92857143],
           [ 4.95238095,  7.57142857],
           [ 7.57142857,  6.02380952],
           [ 7.        ,  2.42857143],
           [ 2.83333333,  2.42857143],
           [ 1.14285714,  4.9047619 ],
           [ 1.35714286,  8.85714286],
           [ 5.52380952,  8.85714286],
           [ 8.85714286,  8.02380952]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')
        _numeric('scan_length', 'floating',
                 range_='[{};inf)'.format(_min_scan_length))
        _numeric('num_points', 'integer',
                 range_='[{};inf)'.format(_min_num_points))

    validate_input()

    coords = square_spiral_sample_surface(float(h - 1), float(w - 1),
                                          scan_length, float(num_points), 1.0)
    coords = coords + 0.5

    return coords


def square_spiral_sample_surface(l, w, speed, sample_rate, time):
    """
    Sample a surface area using a square spiral pattern.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using a square spiral pattern are determined. The
    scanned path is determined from the probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import square_spiral_sample_surface
    >>> l = 1e-6
    >>> w = 1e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> np.set_printoptions(suppress=True)
    >>> square_spiral_sample_surface(l, w, speed, sample_rate, time)
    array([[ 0.0000005,  0.0000005],
           [ 0.0000004,  0.0000004],
           [ 0.0000006,  0.0000007],
           [ 0.0000005,  0.0000003],
           [ 0.0000002,  0.0000007],
           [ 0.0000008,  0.0000007],
           [ 0.0000006,  0.0000002],
           [ 0.0000001,  0.0000004],
           [ 0.0000003,  0.0000009],
           [ 0.0000009,  0.0000008],
           [ 0.0000009,  0.0000001],
           [ 0.0000002,  0.0000001]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('l', 'floating', range_='[{};inf)'.format(_min_l))
        _numeric('w', 'floating', range_='[{};inf)'.format(_min_w))
        _numeric('speed', 'floating', range_='[{};inf)'.format(_min_speed))
        _numeric('sample_rate', 'floating',
                 range_='[{};inf)'.format(_min_sample_rate))
        _numeric('time', 'floating', range_='[{};inf)'.format(_min_time))

    validate_input()

    samples = int(sample_rate * time)
    scan_length = speed * time
    sample_dist = scan_length / samples
    min_size = np.min([l, w])
    delta_dist = np.abs(l - w) / 2

    # Determine number of rotations and length of line segments in spiral
    roots = np.polynomial.polynomial.polyroots(
        [-(scan_length + 2 * delta_dist),
         min_size + 2 * delta_dist - scan_length,
         min_size + 2 * delta_dist])  # [c, b, a] in ax^2 + bx + c = 0
    N = np.ceil(roots.max())  # scan_length ~= (1+1+2+2+..+N+N) * line_length
    line_len = min_size / (N + 1)

    coords = np.zeros((samples, 2))
    position = np.zeros(2)  # Start spiral at (0, 0); later shift its position
    directions = np.array([[-1, 0], [0, 1], [1, 0], [0, -1]])  # -x, +y, +x, -y
    direction_num = 0  # Start spiral by a horisontal line to the left

    # Handle possibly non-square sampling area
    delta = delta_dist * np.array([w > l, w < l])

    n = 1  # n \in (range(N) + 1)
    k = 0  # k = 0 <-> horisontal line segment; k = 1 <-> vertical line segment
    for sample in range(samples):
        # Mark coordinates
        coords[sample, :] = position

        # Update position
        position = position + directions[direction_num, :] * sample_dist

        # Handle corners
        if np.abs(position[k]) > np.ceil(n / 2) * line_len + delta[k]:
            # Fix overshoot from last iteration
            wrap_dist = np.abs(position[k]) - (np.ceil(n / 2) * line_len +
                                               delta[k])
            position = position - directions[direction_num, :] * wrap_dist

            # Find offset if several corners are involved
            offset = np.zeros(2)
            while wrap_dist > n * line_len + 2 * delta[k - 1]:
                wrap_dist = wrap_dist - n * line_len + 2 * delta[k - 1]
                direction_num = np.mod(direction_num + 1, 4)
                k = np.mod(direction_num, 2)
                if k == 0:
                    n = n + 1
                offset = offset + ((n * line_len + 2 * delta[k]) *
                                   directions[direction_num, :])
            else:
                direction_num = np.mod(direction_num + 1, 4)
                k = np.mod(direction_num, 2)
                if k == 0:
                    n = n + 1

            position = position + offset
            position = position + directions[direction_num, :] * wrap_dist

    # Shift to center of region
    coords = coords + np.array([w / 2, l / 2])

    return coords



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing input/output functionality for .mi files.

Currently, only input functionality is provided.

Routine listings
----------------
read_mi_file(path)
    Read an .mi file given by a path.

"""

from magni.afm.io._util import read_mi_file



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for building a hierarchical object-structure.

The reading of an .mi file is logically separated into four steps of which the
functionality provided by this module performs the second step.

Routine listings
----------------
build_object(params)
    Build a hierarchical object-structure from header parameters.

"""

from __future__ import division
from collections import OrderedDict as _OrderedDict

from magni.afm.types.spectroscopy import Chunk as _Chunk
from magni.afm.types.spectroscopy import Grid as _Grid
from magni.afm.types.spectroscopy import Point as _Point


def build_object(params):
    """
    Build a hierarchical object-structure from header parameters.

    The hierarchical object-structure mimics that of `magni.afm.types`.

    Parameters
    ----------
    params : list or tuple
        The header parameters.

    Returns
    -------
    obj : object
        A hierarchical object-structure.

    See Also
    --------
    magni.afm.io.read_mi_file : Function using the present function.

    Notes
    -----
    This function splits the header parameters into file-related parameters and
    buffer-related parameters.

    """

    groups = [[]]

    for name, value in params:
        if name == 'bufferLabel':
            groups.append([])

        groups[-1].append((name, value))

    groups[0].append(groups[-1].pop())
    attrs = dict(groups[0])

    for name in ('xPixels', 'yPixels'):
        if name not in attrs.keys():
            raise IOError('The file must have a {!r} header parameter.'
                          .format(name))

        if not isinstance(attrs[name], int):
            raise IOError('The {!r} file header must have type {!r}.'
                          .format(name, int))

    buffers = [_build_buffer(attrs['fileType'], index, params_)
               for index, params_ in enumerate(groups[1:])]

    if len(buffers) == 0:
        raise IOError("The file must have at least one 'bufferLabel' header.")

    if attrs['fileType'].lower() == 'spectroscopy':
        _expand_buffers(buffers)

    obj = {'attrs': attrs, 'buffers': buffers}
    return _handle_format_inconsistency(obj)


def _build_buffer(file_type, index, params):
    """
    Build a buffer-like object-structure from buffer parameters.

    For spectroscopy buffers, this function converts parameters which contain
    "sub-parameters" to objects with attributes.

    Parameters
    ----------
    file_type : str
        The .mi file type.
    index : int
        The index of the buffer.
    params : list or tuple
        The buffer parameters.

    Returns
    -------
    obj : object
        The buffer-like object-structure.

    See Also
    --------
    build_object : Function using the present function.

    """

    if file_type.lower() == 'image':
        buffer_ = {'attrs': dict(params)}
    else:  # 'spectroscopy'
        allowed_params = ('chunk',) if index > 0 else ('grid', 'point')
        multiparams = {'grid': {'types': _Grid.params, 'min_len': 9},
                       'point': {'types': _Point.params, 'min_len': 3},
                       'chunk': {'types': _Chunk.params, 'min_len': 7}}

        # grid xDirection and yDirection are stored as integers
        multiparams['grid']['types'] = _OrderedDict(_Grid.params.items())
        multiparams['grid']['types']['xDirection'] = int
        multiparams['grid']['types']['yDirection'] = int

        attrs = []
        data = []

        for name, value in params:
            if name in multiparams.keys():
                if name not in allowed_params:
                    raise IOError('Buffer #{!r} may not contain {!r} headers.'
                                  .format(index, name))

                types = multiparams[name]['types']
                min_len = multiparams[name]['min_len']
                params = []

                if len(value) < min_len:
                    msg = ('Each {!r} buffer header must contain at least '
                           '{!r} values.')
                    raise IOError(msg.format(name, len(types)))

                for value, param, type_ in zip(value, *zip(*types.items())):
                    if not isinstance(value, type_):
                        msg = ('The values of each {!r} buffer header must '
                               'have types {!r}.')
                        raise IOError(msg.format(name, types))

                    params.append((param, value))

                data.append({'type': name, 'attrs': dict(params)})
            else:
                attrs.append((name, value))

        buffer_ = {'attrs': dict(attrs), 'data': data}

    return buffer_


def _expand_buffers(buffers):
    """
    Expand the buffer-like object-structures.

    Grid parameters specify implicit point parameters. Furthermore, some chunk
    parameters may be implicitly specified. These implicit parameters should be
    made explicit.

    Parameters
    ----------
    buffers : list or tuple
        The buffer-like object-structures.

    See Also
    --------
    build_object : Function using the present function.
    _generate_grid_points : Make implicit point parameters explicit.
    _generate_implicit_chunks : Make implicit chunk parameters explicit.

    Notes
    -----
    This function relies on the two functions, `_generate_grid_points` and
    `_generate_implicit_chunks` to make the implicit parameters explicit.

    """

    indices = {'grid': 0, 'point': 0}

    for i, item in enumerate(buffers[0]['data']):
        if item['attrs']['index'] != indices[item['type']]:
            raise IOError('The {!r} buffer parameter indices are erroneous.'
                          .format(item['type']))

        if item['type'] == 'grid':
            try:
                attr = 'xDirection'
                item['attrs'][attr] = ('trace', 'retrace')[item['attrs'][attr]]
                attr = 'yDirection'
                item['attrs'][attr] = ('up', 'down')[item['attrs'][attr]]
            except IndexError:
                msg = "The {!r} of 'grid' buffer parameters must be in {!r}."
                raise IOError(msg.format(attr, (0, 1)))

            points = _generate_grid_points(item['attrs'], indices['point'])
            item['points'] = points
            indices['point'] = indices['point'] + len(points) * len(points[0])

        indices[item['type']] = indices[item['type']] + 1

    for buffer_ in buffers[1:]:
        buffer_['data'] = _generate_implicit_chunks(indices['point'],
                                                    buffer_['data'])


def _generate_grid_points(attrs, index):
    """
    Make implicit grid point parameters explicit.

    Parameters
    ----------
    attrs : list or tuple
        The attributes of the grid.
    index : int
        The index of the first grid point.

    Returns
    -------
    points : tuple
        The grid points.

    See Also
    --------
    _expand_buffers : Function using the present function.

    """

    points = []
    spacing = attrs['pointSpacing']
    x_start = attrs['xCenter'] - (attrs['xPoints'] - 1) / 2 * spacing
    y_start = attrs['yCenter'] - (attrs['yPoints'] - 1) / 2 * spacing

    x_dir = slice(None, None, 1 if attrs['xDirection'] == 'trace' else -1)
    y_dir = slice(None, None, 1 if attrs['yDirection'] == 'down' else -1)

    for y_index in range(attrs['yPoints']).__getitem__(y_dir):
        points.append([])
        y = y_start + y_index * spacing

        for x_index in range(attrs['xPoints']).__getitem__(x_dir):
            x = x_start + x_index * spacing
            point = {'index': index, 'xCoordinate': x, 'yCoordinate': y}
            points[-1].append({'attrs': point})
            index = index + 1

        points[-1] = points[-1].__getitem__(x_dir)

    points = points.__getitem__(y_dir)
    return points


def _generate_implicit_chunks(npoints, explicit_chunks):
    """
    Make implicit chunk parameters explicit.

    Parameters
    ----------
    npoints : int
        The number of points.
    explicit_chunks : list
        The explicit chunks.

    Returns
    -------
    chunks : list
        The chunks including both explicit and implicit chunks.

    See Also
    --------
    _expand_buffer : Function using the present function.

    """

    chunks = []
    point_index = 0
    point_chunks = []
    stop_chunk = {'attrs': {'pointIndex': npoints, 'label': '_emptychunk_'}}

    for chunk in explicit_chunks + [stop_chunk]:
        index = chunk['attrs'].get('pointIndex', 0)

        for point_index in range(point_index + 1, index):
            for point_chunk in point_chunks:
                point_chunk = point_chunk.copy()
                point_chunk['attrs'] = point_chunk['attrs'].copy()
                point_chunk['attrs']['pointIndex'] = point_index
                chunks.append(point_chunk)

        if (chunk['attrs'].get('label', '').lower() == '_emptychunk_' and
                index <= npoints):
            break

        if point_index < index < npoints:
            point_index = point_index + 1
            point_chunks = []

        if index != point_index:
            raise IOError("The 'chunk' buffer parameter indices are "
                          "erroneous.")

        chunks.append(chunk)
        point_chunks.append(chunk)

    return chunks


def _handle_format_inconsistency(obj):
    """
    Handle format inconsistency.

    The inconsistency is the usage of string values for the 'trace' header
    parameter which is specified to have a boolean value.

    Parameters
    ----------
    obj : object
        The hierarchical object-structure.

    Returns
    -------
    obj : object
        The updated hierarchical object-structure.

    See Also
    --------
    build_object : Function using the present function.

    """

    if obj['attrs']['fileType'].lower() == 'image':
        for buffer_ in obj['buffers']:
            if ('trace' in buffer_['attrs'].keys() and
                    isinstance(buffer_['attrs']['trace'], str)):
                if buffer_['attrs']['trace'].lower() == 'trace':
                    buffer_['attrs']['trace'] = True
                elif buffer_['attrs']['trace'].lower() == 'retrace':
                    buffer_['attrs']['trace'] = False

    return obj



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the process function.

Routine listings
----------------
process(func, namespace={}, args_list=None, kwargs_list=None, maxtasks=None)
    Map multiple function calls to multiple processors.

See Also
--------
magni.utils.multiprocessing.config : Configuration options.

"""

from __future__ import division
import multiprocessing as mp
import os
import traceback
import warnings

import numpy as np

try:
    # The concurrent.futures was first added to Python in version 3.2
    # A backport for Python 2 is available at https://pythonhosted.org/futures/
    import concurrent.futures
    from concurrent.futures import ProcessPoolExecutor
    from concurrent.futures.process import BrokenProcessPool
    _futures_available = True
except ImportError:
    _futures_available = False

try:
    # disable mkl multiprocessing to avoid conflicts with manual
    # multiprocessing
    import mkl
    _get_num_threads = mkl.get_max_threads
    _set_num_threads = mkl.set_num_threads
except ImportError:
    def _get_num_threads():
        return 0

    def _set_num_threads(n):
        pass

from magni.utils.multiprocessing import config as _config
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


def process(func, namespace={}, args_list=None, kwargs_list=None,
            maxtasks=None):
    """
    Map multiple function calls to multiple processes.

    For each entry in args_list and kwargs_list, a task is formed which is used
    for a function call of the type `func(*args, **kwargs)`. Each task is
    executed in a seperate process using the concept of a processing pool.

    Parameters
    ----------
    func : function
        A function handle to the function which the calls should be mapped to.
    namespace : dict, optional
        A dict whose keys and values should be globally available in func (the
        default is an empty dict).
    args_list : list or tuple, optional
        A sequence of argument lists for the function calls (the default is
        None, which implies that no arguments are used in the calls).
    kwargs_list : list or tuple, optional
        A sequence of keyword argument dicts for the function calls (the
        default is None, which implies that no keyword arguments are used in
        the calls).
    maxtasks : int, optional
        The maximum number of tasks of a process before it is replaced by a new
        process (the default is None, which implies that processes are not
        replaced).

    Returns
    -------
    results : list
        A list with the results from the function calls.

    Raises
    ------
    BrokenPoolError
        If using the `concurrent.futures` module and one or more workers
        terminate abrubtly with the automatic broken pool restart funtionality
        disabled.

    See Also
    --------
    magni.utils.multiprocessing.config : Configuration options.

    Notes
    -----
    If the `workers` configuration option is equal to 0, map is used.
    Otherwise, the map functionality of a processing pool is used.

    Reasons for using this function over map or standard multiprocessing:

    - Simplicity of the code over standard multiprocessing.
    - Simplicity in switching between single- and multiprocessing.
    - The use of both arguments and keyword arguments in the function calls.
    - The reporting of exceptions before termination.
    - The possibility of terminating multiprocessing with a single interrupt.
    - The option of automatically restarting a broken process pool.

    As of Python 3.2, two different, though quite similar, modules exist in the
    standard library for managing processing pools: `multiprocessing` and
    `concurrent.futures`. According to Python core contributor Jesse Noller,
    the plan is to eventually make concurrent.futures the only interface to the
    high level processing pools (futures), whereas multiprocessing is supposed
    to serve more low level needs for individually handling processes, locks,
    queues, etc. (see https://bugs.python.org/issue9205#msg132661). As of
    Python 3.5, both the `multiprocessing.Pool` and
    `concurrent.futures.ProcessPoolExecutor` serve almost the same purpose and
    provide very similar interfaces. The main differences between the two are:

    - The option of using a worker initialiser is only available in
      `multiprocessing`.
    - The option of specifing a maximum number of tasks for a worker to execute
      before being replaced to free up ressources (the maxtasksperchild option)
      is only available in `multiprocessing`.
    - The option of specifying a context is only available in `multiprocessing`
    - "Reasonable" handling of abrubt worker termination and exceptions is only
       available in `concurrent.futures`.

    Particularly, the "reasonable" handling of a broken process pool may be a
    strong argument to prefer `concurrent.futures` over `multiprocessing`. The
    matter of handling a broken process pool has been extensively discussed in
    https://bugs.python.org/issue9205 which led to the fix for
    `concurrent.futures`. A similar fix for `multiprocessing` has been proposed
    in https://bugs.python.org/issue22393.

    Both the `multiprocessing` and `concurrent.futures` interfaces are
    available for use with this function. If the configuration parameter
    `prefer_futures` is set to True and the `concurrent.futures` module is
    available, this is used. Otherwise, the `multiprocessing` module is used. A
    Python 2 backport of `concurrent.futures` is available at
    https://pythonhosted.org/futures/.

    When using `concurrent.futures`, the `namespace` and `init_args` are
    ignored since these are not supported by that module. Support for
    `maxtasks` is emulated in a way such that each process on average does
    `maxtasks` before it is restarted. The `init_args` functionality may be
    added later on if an initialiser is added to `concurrent.futures` - see
    http://bugs.python.org/issue21423. If the `max_broken_pool_restarts`
    configuration parameter is set to a value different from 0, the Pool is
    automatically restarted and the tasks are re-run should a broken pool be
    encountered. If `max_broken_pool_restarts` is set to 0, a BrokenPoolError
    is raised should a broken pool be encountered.

    When using `multiprocessing`, the `max_broken_pool_restarts` is ignored
    since the BrokenPoolError handling has not yet been implemented for the
    `multiprocessing.Pool` - see https://bugs.python.org/issue22393 as well as
    https://bugs.python.org/issue9205.

    Examples
    --------
    An example of how to use args_list and kwargs_list:

    >>> import magni
    >>> from magni.utils.multiprocessing._processing import process
    >>> def calculate(a, b, op='+'):
    ...     if op == '+':
    ...         return a + b
    ...     elif op == '-':
    ...         return a - b
    ...
    >>> args_list = [[5, 7], [9, 3]]
    >>> kwargs_list = [{'op': '+'}, {'op': '-'}]
    >>> process(calculate, args_list=args_list, kwargs_list=kwargs_list)
    [12, 6]

    or the same example preferring `concurrent.futures` over `multiprocessing`:

    >>> magni.utils.multiprocessing.config['prefer_futures'] = True
    >>> process(calculate, args_list=args_list, kwargs_list=kwargs_list)
    [12, 6]

    """

    @_decorate_validation
    def validate_input():
        _generic('func', 'function')
        _generic('namespace', 'mapping')
        _levels('args_list', (_generic(None, 'collection', ignore_none=True),
                              _generic(None, 'explicit collection')))
        _levels('kwargs_list', (_generic(None, 'collection', ignore_none=True),
                                _generic(None, 'mapping')))

        if args_list is None and kwargs_list is None:
            msg = ('The value of >>args_list<<, {!r}, and/or the value of '
                   '>>kwargs_list<<, {!r}, must be different from {!r}.')
            raise ValueError(msg.format(args_list, kwargs_list, None))
        elif args_list is not None and kwargs_list is not None:
            if len(args_list) != len(kwargs_list):
                msg = ('The value of >>len(args_list)<<, {!r}, must be equal '
                       'to the value of >>len(kwargs_list)<<, {!r}.')
                raise ValueError(msg.format(len(args_list), len(kwargs_list)))

        _numeric('maxtasks', 'integer', range_='(0;inf)', ignore_none=True)

    validate_input()

    tasks = _get_tasks(func, args_list, kwargs_list)

    if _config['workers'] == 0:
        _process_init(func, namespace)
        results = list(map(_process_worker, tasks))
    else:
        if os.name == 'nt':
            raise NotImplementedError('This function is not available under '
                                      'Windows.')

        if _futures_available and _config['prefer_futures']:
            map_ = _map_using_futures
        else:
            map_ = _map_using_mppool

        try:
            num_threads = _get_num_threads()
            _set_num_threads(1)
            results = map_(_process_worker, tasks, (func, namespace), maxtasks,
                           _config['max_broken_pool_restarts'])
        finally:
            _set_num_threads(num_threads)

    return results


def _get_tasks(func, args_list, kwargs_list):
    """
    Prepare a list of tasks.

    Parameters
    ----------
    func : function
        A function handle to the function which the calls should be mapped to.
    args_list : list or tuple, optional
        A sequence of argument lists for the function calls (the default is
        None, which implies that no arguments are used in the calls).
    kwargs_list : list or tuple, optional
        A sequence of keyword argument dicts for the function calls (the
        default is None, which implies that no keyword arguments are used in
        the calls).

    Returns
    -------
    tasks : list
        The list of tasks.

    """

    if args_list is None:
        args_list = [() for dct in kwargs_list]

    if kwargs_list is None:
        kwargs_list = [{} for lst in args_list]

    tasks = [func for lst in args_list]
    tasks = list(zip(tasks, args_list, kwargs_list))

    return tasks


def _map_using_futures(func, tasks, init_args, maxtasks,
                       max_broken_pool_restarts):
    """
    Map a set of `tasks` to `func` and run them in parallel using a futures.

    If `max_broken_pool_restarts` is different from 0, the tasks must be an
    explicit collection, e.g. a list or tuple, for the restart to work. If an
    exception occurs in one of the function calls, the process pool terminates
    ASAP and re-raises the first exception that occurred. All exceptions that
    may have occurred in the workers are available as the last element in the
    exceptions args.

    Parameters
    ----------
    func : function
        A function handle to the function which the calls should be mapped to.
    tasks : iterable
        The list of tasks to use as arguments in the function calls.
    init_args : tuple
        The (func, namespace) tuple for the _process_init initialisation
        function.
    maxtasks : int
        The maximum number of tasks of a process before it is replaced by a new
        process. If set to None, the process is never replaced.
    max_broken_pool_restarts : int or None
        The maximum number of attempts at restarting the process pool upon a
        BrokenPoolError. If set to None, the process pool may restart
        indefinitely.

    Returns
    -------
    results : list
        The list of results from the map operation.

    Notes
    -----
    The `namespace` and `init_args` are ignored since these are not supported
    by `concurrent.futures`. Support for `maxtasks` is emulated in a way such
    that each process on average does `maxtasks` before it is restarted. The
    `init_args` functionality may be added if an initialiser is added to
    `concurrent.futures` - see http://bugs.python.org/issue21423.

    """

    num_workers = _config['workers']

    def _fix_none_maxtasks(maxtasks):
        """Handle maxtasks==None correctly."""
        if maxtasks is None:
            return int(len(tasks) / num_workers)
        else:
            return maxtasks
    maxtasks = _fix_none_maxtasks(maxtasks)

    try:
        batch_size = num_workers * maxtasks
        future_batches = max(int(np.ceil(len(tasks) / batch_size)), 1)
        future_results = []

        for l in range(future_batches):
            # Restart ProcessPool when all workers have done `maxtasks` on
            # average in order to free ressources
            workers = ProcessPoolExecutor(max_workers=num_workers)
            futures = [workers.submit(func, task)
                       for task in tasks[l*batch_size:(l+1)*batch_size]]
            concurrent.futures.wait(
                futures, return_when=concurrent.futures.FIRST_EXCEPTION)
            future_results.extend([future.result() for future in futures])
            workers.shutdown()

        return future_results

    except BrokenProcessPool as e:
        workers.shutdown()
        base_msg = 'A BrokenProcessPool was encountered. '

        if max_broken_pool_restarts is None:
            msg = base_msg + 'Restarting the process pool.'
            warnings.warn(msg, RuntimeWarning, stacklevel=2)
            return _map_using_futures(func, tasks, init_args, maxtasks, None)
        elif max_broken_pool_restarts > 0:
            new_restart_count = max_broken_pool_restarts - 1
            msg = (base_msg + 'Restarting the process pool with ' +
                   'max_broken_pool_restarts={}.').format(new_restart_count)
            warnings.warn(msg, RuntimeWarning, stacklevel=2)
            return _map_using_futures(
                func, tasks, init_args, maxtasks, new_restart_count)
        else:
            msg = (base_msg + 'Giving up on restarting the process pool ' +
                   'and re-raising.')
            warnings.warn(msg, RuntimeWarning, stacklevel=2)
            raise
    except BaseException as e:
        worker_exceptions = [future.exception() for future in futures]
        msg = ('An exception occurred in one or more workers. ' +
               'Re-raising with all exceptions appended to the current ' +
               'exceptions arguments.')
        warnings.warn(msg, RuntimeWarning)
        e.args = e.args + tuple(worker_exceptions)
        raise
    finally:
        workers.shutdown()


def _map_using_mppool(func, tasks, init_args, maxtasks,
                      max_broken_pool_restarts):
    """
    Map a set of `tasks` to `func` and run them in parallel via multiprocessing

    Parameters
    ----------
    func : function
        A function handle to the function which the calls should be mapped to.
    tasks : iterable
        The list of tasks to use as arguments in the function calls.
    init_args : tuple
        The (func, namespace) tuple for the _process_init initialisation
        function.
    maxtasks : int
        The maximum number of tasks of a process before it is replaced by a new
        process. If set to None, the process is never replaced.
    max_broken_pool_restarts : int or None
        The maximum number of attempts at restarting the process pool upon a
        BrokenPoolError. If set to None, the process pool may restart
        indefinitely.

    Returns
    -------
    results : list
        The list of results from the map operation.

    Notes
    -----
    The `max_broken_pool_restarts` is ignored since the BrokenPoolError
    handling has not yet been implemented in the multiprocessing.Pool - see
    https://bugs.python.org/issue22393 and https://bugs.python.org/issue9205.

    """

    try:
        workers = mp.Pool(
            _config['workers'], _process_init, init_args, maxtasks)
        return workers.map(func, tasks, chunksize=1)
    finally:
        workers.close()
        workers.join()


def _process_init(func, namespace):
    """
    Initialise the process by making global variables available to it.

    Parameters
    ----------
    func : function
        A function handle to the function which the calls should be mapped to.
    namespace : dict
        A dict whose keys and values should be globally available in func.

    """

    func.__globals__.update(namespace)


def _process_worker(fak_tuple):
    """
    Unpack and map a task to the function.

    Parameters
    ----------
    fak_tuple : tuple
        A tuple (func, args, kwargs) containing the parameters listed below.
    func : function
        A function handle to the function which the calls should be mapped to.
    args : list or tuple
        The sequence of arguments that should be unpacked and passed.
    kwargs : list or tuple
        The sequence of keyword arguments that should be unpacked and passed.

    Notes
    -----
    If an exception is raised in `func`, the stacktrace of that exception is
    printed since the exception is otherwise silenced until every task has been
    executed when using multiple workers.

    Also, a workaround has been implemented to allow KeyboardInterrupts to
    interrupt the current tasks and all remaining tasks. This is done by
    setting a global variable, when catching a KeyboardInterrupt, which is
    checked for every call.

    """

    func, args, kwargs = fak_tuple

    if 'interrupted' not in _process_worker.__globals__:
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            _process_worker.__globals__['interrupted'] = True
        except BaseException as e:
            print(traceback.format_exc())

            if _config['silence_exceptions']:
                return e
            elif _config['re_raise_exceptions']:
                raise
            else:
                raise RuntimeError('An exception has occured.')



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
uniform_rotated_line_sample_image(h, w, scan_length, num_points, angle=0.,
    follow_edge=True)
    Function for uniform rotated line sampling an image.
uniform_rotated_line_sample_surface(l, w, speed, sample_rate, time, angle=0.,
    follow_edge=True)
    Function for uniform rotated line sampling a surface.

"""

from __future__ import division

import numpy as np

from magni.imaging.measurements import _util
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['uniform_rotated_line_sample_image',
           'uniform_rotated_line_sample_surface']

_min_l = _util.min_l
_min_w = _util.min_w
_min_speed = _util.min_speed
_min_sample_rate = _util.min_sample_rate
_min_time = _util.min_time
_min_scan_length = _util.min_scan_length
_min_num_points = _util.min_num_points


def uniform_rotated_line_sample_image(h, w, scan_length, num_points, angle=0.,
                                      follow_edge=True):
    r"""
    Sample an image using a uniform rotated line pattern.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a uniform rotated line pattern are determined. The
    `scan_length` determines the length of the path scanned whereas
    `num_points` indicates the number of samples taken on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.
    angle : float
        The angle measured in radians by which the uniform lines are rotated
        (the default is 0.0 resulting in a pattern identical to that of
        uniform_line_sample_image).
    follow_edge: bool
        A flag indicating whether or not the pattern follows the edges of the
        rectangular area in-between lines (the default is True).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    The `angle` is limited to the interval :math:`[0;\pi)`. An `angle` of 0
    results in the same behaviour as that of uniform_line_sample_image. An
    increase in the `angle` rotates the overall direction counterclockwise,
    i.e., at :math:`\frac{\pi}{2}`, the uniform_line_sample_image sampling
    pattern is rotated 90 degrees counterclockwise.

    If the `follow_edge` flag is True, then the pattern follows the edges of
    the rectangular area when moving from one line to the next. If the flag is
    False, then the pattern follows a line perpendicular to the uniform lines
    when moving from one line to the next. In the latter case, some of the
    uniform lines are shortened to allow the described behaviour.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import \
    ... uniform_rotated_line_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> np.set_printoptions(suppress=True)
    >>> uniform_rotated_line_sample_image(h, w, scan_length, num_points)
    array([[ 0.5       ,  0.5       ],
           [ 4.59090909,  0.5       ],
           [ 8.68181818,  0.5       ],
           [ 9.22727273,  3.5       ],
           [ 5.13636364,  3.5       ],
           [ 1.04545455,  3.5       ],
           [ 1.04545455,  6.5       ],
           [ 5.13636364,  6.5       ],
           [ 9.22727273,  6.5       ],
           [ 8.68181818,  9.5       ],
           [ 4.59090909,  9.5       ],
           [ 0.5       ,  9.5       ]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')
        _numeric('scan_length', 'floating',
                 range_='[{};inf)'.format(_min_scan_length))
        _numeric('num_points', 'integer',
                 range_='[{};inf)'.format(_min_num_points))
        _numeric('angle', 'floating', range_='[0;{})'.format(np.pi))
        _numeric('follow_edge', 'boolean')

    validate_input()

    coords = uniform_rotated_line_sample_surface(
        float(h - 1), float(w - 1), scan_length, float(num_points - 1), 1.,
        angle, follow_edge)
    coords = coords + 0.5

    return coords


def uniform_rotated_line_sample_surface(l, w, speed, sample_rate, time,
                                        angle=0., follow_edge=True):
    r"""
    Sample a surface area using a uniform rotated line pattern.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using uniform rotated line pattern are determined. The
    scanned path is determined from the probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.
    angle : float
        The angle measured in radians by which the uniform lines are rotated
        (the default is 0.0 resulting in a pattern identical to that of
        uniform_line_sample_image).
    follow_edge: bool
        A flag indicating whether or not the pattern foolows the edges of the
        rectangular area in-between lines (the default is True).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis.

    The `angle` is limited to the interval :math:`[0;\pi)`. An `angle` of 0
    results in the same behaviour as that of uniform_line_sample_surface. An
    increase in the `angle` rotates the overall direction counterclockwise,
    i.e., at :math:`\frac{\pi}{2}`, the uniform_line_sample_surface sampling
    pattern is rotated 90 degrees counterclockwise.

    If the `follow_edge` flag is True, then the pattern follows the edges of
    the rectangular area when moving from one line to the next. If the flag is
    False, then the pattern follows a line perpendicular to the uniform lines
    when moving from one line to the next. In the latter case, some of the
    uniform lines are shortened to allow the described behaviour.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import \
    ... uniform_rotated_line_sample_surface
    >>> l = 1e-6
    >>> w = 1e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> np.set_printoptions(suppress=True)
    >>> uniform_rotated_line_sample_surface(l, w, speed, sample_rate, time)
    array([[ 0.        ,  0.        ],
           [ 0.00000067,  0.        ],
           [ 0.00000083,  0.00000017],
           [ 0.00000017,  0.00000017],
           [ 0.00000033,  0.00000033],
           [ 0.000001  ,  0.00000033],
           [ 0.0000005 ,  0.0000005 ],
           [ 0.        ,  0.00000067],
           [ 0.00000067,  0.00000067],
           [ 0.00000083,  0.00000083],
           [ 0.00000017,  0.00000083],
           [ 0.00000033,  0.000001  ],
           [ 0.000001  ,  0.000001  ]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('l', 'floating', range_='[{};inf)'.format(_min_l))
        _numeric('w', 'floating', range_='[{};inf)'.format(_min_w))
        _numeric('speed', 'floating', range_='[{};inf)'.format(_min_speed))
        _numeric('sample_rate', 'floating',
                 range_='[{};inf)'.format(_min_sample_rate))
        _numeric('time', 'floating', range_='[{};inf)'.format(_min_time))
        _numeric('angle', 'floating', range_='[0;{})'.format(np.pi))
        _numeric('follow_edge', 'boolean')

    validate_input()

    if angle >= np.pi / 2:
        l, w = w, l
        angle = angle - np.pi / 2
        rotate = True
    else:
        rotate = False

    cos = np.cos(angle)
    sin = np.sin(angle)

    if angle > 0:
        coords_corner = np.float_([[cos * w - sin * l, cos * l + sin * w]])
        dir_l = -sin * coords_corner[0, 1] / cos
        dir_w = cos * coords_corner[0, 1] / sin

    def transform(coords):
        coords[:, 1] = -coords[:, 1]
        coords = coords.T
        coords = np.float_([[cos, -sin], [sin, cos]]).dot(coords)
        coords = coords.T
        coords[:, 1] = -coords[:, 1]
        return coords

    n = 3
    coords_prev = np.zeros((3, 2))

    while True:
        ratios = np.linspace(0, 1, n).reshape((n, 1))

        if angle > 0:
            Y = ratios * coords_corner[0, 1]
            X_lower = np.maximum(dir_l * ratios,
                                 coords_corner[0, 0] - dir_w * (1 - ratios))
            X_upper = np.minimum(dir_w * ratios,
                                 coords_corner[0, 0] - dir_l * (1 - ratios))
        else:
            Y = ratios * l
            X_lower = 0 + ratios * 0
            X_upper = w + ratios * 0

        coords_lower = np.column_stack((X_lower, Y))
        coords_upper = np.column_stack((X_upper, Y))

        if follow_edge:
            coords_lower = transform(coords_lower)
            coords_upper = transform(coords_upper)
            coords = np.zeros((3 * n, 2))

            # alternate between points
            coords[1::6] = coords_lower[0::2]
            coords[2::6] = coords_upper[0::2]
            coords[4::6] = coords_upper[1::2]
            coords[5::6] = coords_lower[1::2]

            # fill the blanks
            coords[3:-1:6, 0] = coords[4:-1:6, 0]
            coords[3:-1:6, 1] = coords[2:-1:6, 1]
            coords[6:-1:6, 0] = coords[5:-1:6, 0]
            coords[6:-1:6, 1] = coords[7:-1:6, 1]
        else:
            coords = np.zeros((2 * n, 2))

            # alternate between points
            coords[0::4] = coords_lower[0::2]
            coords[1::4] = coords_upper[0::2]
            coords[2::4] = coords_upper[1::2]
            coords[3::4] = coords_lower[1::2]

            # shorten
            coords[1:-1:4, 0] = coords[2:-1:4, 0] = np.minimum(
                coords[1:-1:4, 0], coords[2:-1:4, 0])
            coords[3:-1:4, 0] = coords[4:-1:4, 0] = np.maximum(
                coords[3:-1:4, 0], coords[4:-1:4, 0])

            coords = transform(coords)

        length = coords[1:] - coords[:-1]
        length = np.sum(np.sqrt(length[:, 0]**2 + length[:, 1]**2))

        if length > speed * time:
            n = n - 1
            coords = coords_prev
            break
        else:
            n = n + 1
            coords_prev = coords

    if rotate:
        l, w = w, l
        coords[:, 0], coords[:, 1] = coords[:, 1], l - coords[:, 0]

    X = coords[:, 0]
    X[X < 0] = 0
    X[X > w] = w
    Y = coords[:, 1]
    Y[Y < 0] = 0
    Y[Y > l] = l

    return _util.sample_lines(coords, speed, sample_rate, time)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for analysing the simulation results.

Routine listings
----------------
run(path, label)
    Determine the phase transition from the simulation results.

See Also
--------
magni.cs.phase_transition.config : Configuration options.

Notes
-----
For a description of the concept of phase transition, see [1]_.

References
----------
.. [1] C. S. Oxvig, P. S. Pedersen, T. Arildsen, and T. Larsen, "Surpassing the
   Theoretical 1-norm Phase Transition in Compressive Sensing by Tuning the
   Smoothed l0 Algorithm", *in IEEE International Conference on Acoustics,
   Speech and Signal Processing (ICASSP)*, Vancouver, Canada, May 26-31, 2013,
   pp. 6019-6023.

"""

from __future__ import division

import numpy as np

from magni.cs.phase_transition import config as _conf
from magni.utils.multiprocessing import File as _File


def run(path, label):
    """
    Determine the phase transition from the simulation results.

    The simulation results should be present in the HDF5 database specified by
    `path` in the pytables group specified by `label` in an array named 'dist'.
    The determined phase transition (50% curve) is stored in the same HDF5
    database, in the same HDF group in an array named 'phase_transition'.
    Additionally, the 10%, 25%, 75%, and 90% percentiles are stored in an array
    named 'phase_transition_percentiles'.

    Parameters
    ----------
    path : str
        The path of the HDF5 database.
    label : str
        The path of the pytables group in the HDF5 database.

    See Also
    --------
    _estimate_PT : The actual phase transition estimation.

    Notes
    -----
    A simulation is considered successful if the simulation result is less than
    a normalised mean squared error tolerance computed as 10^(-SNR/10) wtih SNR
    configured in the configuration module.

    """

    percentiles = [0.5, 0.1, 0.25, 0.75, 0.9]

    with _File(path, 'a') as f:
        if not '/' + label + '/phase_transition' in f:
            dist = f.get_node('/' + label + '/dist')[:]
            # Set NaNs to value > NMSE_tolerance, i.e. assume failure
            dist[np.isnan(dist)] = 10**(-_conf['SNR'] / 10) * 2

            if _conf['logit_solver'] == 'built-in':
                # Use "simple" built-in solver
                rho = _built_in_logit_solver(dist, percentiles)
            elif _conf['logit_solver'] == 'sklearn':
                # Use scikit learn solver
                rho = _sklearn_logit_solver(dist, percentiles)

            f.create_array('/' + label, 'phase_transition', rho[0])
            f.create_array('/' + label, 'phase_transition_percentiles',
                           rho[1:])


def _built_in_logit_solver(dist, percentiles):
    """
    Fit a logistic regression model using the built-in solver.

    Parameters
    ----------
    dist : ndarray
        The simulated signal "distances" in the phase space.
    percentiles : list or tuple
        The percentiles to estimate.

    Returns
    -------
    rho : ndarray
        The "len(percentiles)"-by-"len(delta)" array of estimated phase
        transition rho vectors. The phase transition rho vectors are
        (in order): 50% (the phase transition esitmate), smaller to larger
        percentiles.

    """

    monte_carlo = dist.shape[2]
    points = len(_conf['rho']) * monte_carlo
    NMSE_tolerance = 10**(-_conf['SNR'] / 10)

    z = np.zeros(points)
    y = np.zeros(points)

    rho = np.zeros((len(percentiles), len(_conf['delta'])))

    for i in range(len(_conf['delta'])):
        n = np.round(_conf['delta'][i] * _conf['problem_size'])

        for j in range(len(_conf['rho'])):
            if n > 0:
                var = _conf['rho'][j]
                var = np.round(var * n) / n
            else:
                var = 0.

            for m in range(monte_carlo):
                z[j * monte_carlo + m] = var
                y[j * monte_carlo + m] = dist[i, j, m] < NMSE_tolerance

        rho[:, i] = _estimate_PT(z, y, percentiles)

    return rho


def _estimate_PT(rho, success, percentiles):
    """
    Estimate the phase transition location for a given delta.

    The phase transition location is estimated using logistic regression. The
    algorithm used for this is Newton's method.

    Parameters
    ----------
    rho : ndarray
        The rho values.
    success : ndarray
        The success indicators.
    percentiles : list or tuple
        The percentiles to estimate.

    Returns
    -------
    rho : float
        The estimated phase transition location.

    Notes
    -----
    The function includes a number of non-standard ways of handling numerical
    and convergence related issues. This will be changed in a future version of
    the code.

    """

    points = len(success)

    if success.sum() < 0.5:
        # if none of the simulations were successful
        return [0] * len(percentiles)
    elif success.sum() > points - 0.5:
        # if all of the simulations were successful
        return [1] * len(percentiles)

    y = np.zeros((points + 2, 1))
    y[:points, 0], y[points:, 0] = success, [1, 0]

    # note: z_i in the algorithm is z[i, :]^T in the implementation
    z = np.ones((points + 2, 2))
    z[:points, 1], z[points:, 1] = rho, [0, 1]

    b = np.zeros((2, 1))

    for l in range(100):
        # p = [p_i]^T
        # p_i = exp(b^T z_i) / (1 + exp(b^T z_i))
        # p_i = 1 / (1 + exp(-z_i^T b))
        p = 1 / (1 + np.exp(-z.dot(b)))

        # gradient
        # g = sum([(y_i - p_i) * z_i])
        g = z.T.dot(y - p)

        # Hessian
        # H = -sum([p_i * (1 - p_i) * z_i z_i^T])
        H = -(p * (1 - p) * z).T.dot(z)

        try:
            # step
            s = np.linalg.inv(H).dot(g)
        except np.linalg.LinAlgError:
            # it results in convergence but is hardly a standard solution
            det = H[0, 0] * H[1, 1] - H[0, 1] * H[1, 0]
            det = np.sign(np.sign(det) + 0.5) * 1e-100

            Hinv = np.float64([[H[1, 1], -H[0, 1]], [-H[1, 0], H[0, 0]]]) / det
            s = Hinv.dot(g)

        b = b - s
        # constrained to non-positive since the model would otherwise suggest
        # a better chance of reconstruction for higher rho
        b[1] = min(b[1], 0)

        g_len = np.linalg.norm(g)
        s_len = np.linalg.norm(s)
        b_len = np.linalg.norm(b)

        # g_len < 1e-12 : convergence
        # s_len < 1e-3 : convergence
        # s_len / b_len < 1e-12 : ... in case of large coefficients
        # the last two should probably be replaced by
        #     delta (-b[0] / b[1]) < 1e-12
        # or something like that
        if g_len < 1e-12 or s_len < 1e-3 or s_len / b_len < 1e-12:
            val = np.zeros(len(percentiles))
            for l, p in enumerate(percentiles):
                val[l] = (np.log(p / (1 - p)) - b[0]) / b[1]
                val[l] = max(val[l], 0)
                val[l] = min(val[l], 1)

            return val

    raise RuntimeWarning('analysis.py: phase transition does not converge.')


def _sklearn_logit_solver(dist, percentiles):
    """
    Fit a logistic regression model using the solver from scikit-learn.

    Parameters
    ----------
    dist : ndarray
        The simulated signal "distances" in the phase space.
    percentiles : list or tuple
        The percentiles to estimate.

    Returns
    -------
    rho : ndarray
        The "len(percentiles)"-by-"len(delta)" array of estimated phase
        transition rho vectors. The phase transition rho vectors are
        (in order): 50% (the phase transition esitmate), smaller to larger
        percentiles.

    """

    from sklearn.linear_model import LogisticRegression
    NMSE_tolerance = 10**(-_conf['SNR'] / 10)

    lr = LogisticRegression(C=1e3, fit_intercept=False,
                            random_state=_conf['seed'])
    rho = np.zeros((len(percentiles), len(_conf['delta'])))
    monte_carlo = dist.shape[2]

    for k in range(len(_conf['delta'])):
        successes = (dist[k, :, :] < NMSE_tolerance).reshape(-1)

        if not np.any(successes):
            rho[:, k] = 0.0
        elif np.all(successes):
            rho[:, k] = 1.0
        else:
            X = np.column_stack(
                (np.ones_like(successes),
                 np.repeat(_conf['rho'], monte_carlo)))
            lr.fit(X, successes)
            b = [lr.coef_[0, 0], lr.coef_[0, 1]]

            for l, p in enumerate(percentiles):
                rho[l, k] = (np.log(p / (1 - p)) - b[0]) / b[1]
                rho[l, k] = max(rho[l, k], 0.0)
                rho[l, k] = min(rho[l, k], 1.0)

    return rho



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for calculating stop criteria used in Iterative
Threholding (IT) algorithms.

Routine listings
----------------
calculate_using_mse_convergence(var)
    Calculate stop criterion based on mse convergence.
calculate_using_residual(var)
    Calculate stop criterion based on residual.
calculate_using_residual_measurments_ratio(var)
    Calculate stop criterion based on the residual to measurements ratio.
get_function_handle(method)
    Return a function handle to a given calculation method.

"""

from __future__ import division

import numpy as np


def wrap_calculate_using_mse_convergence(var):
    """
    Arguments wrapper for `calculate_using_mse_convergence`.

    Calculate stop criterion based on mse convergence.

    """

    n = var['A'].shape[1]
    tolerance = var['tolerance']

    def calculate_using_mse_convergence(var):
        """
        Calculate stop criterion based on mse convergence.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in calculating of the stop criterion.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        mse : float
            The current convergence mean squared error.

        Notes
        -----
        The IT algorithm should converge to a fixed point. This criterion
        is based on the mean squared error of the difference between the
        proposed solution in this iteration and the proposed solution in the
        previous solution.

        """

        mse = 1/n * np.linalg.norm(var['alpha_prev'] - var['alpha'])**2
        stop = mse < tolerance

        return stop, mse

    return calculate_using_mse_convergence


def wrap_calculate_using_residual(var):
    """
    Arguments wrapper for `calculate_using_residual`.

    Calculate stop criterion based on residual.

    """

    tolerance = var['tolerance']
    m = var['A'].shape[0]

    def calculate_using_residual(var):
        """
        Calculate stop criterion based on residual.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in calculating of the stop criterion.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        r_mse : float
            The current mean squared error of the residual.

        Notes
        -----
        This stopping criterion is based on the mean sqaured error of the
        residual.

        """

        r_mse = 1/m * np.linalg.norm(var['r'])**2
        stop = r_mse < tolerance

        return stop, r_mse

    return calculate_using_residual


def wrap_calculate_using_residual_measurements_ratio(var):
    """
    Arguments wrapper for `calculate_using_residual_measurements_ratio`.

    Calculate stop criterion based on the residual to measurements ratio.

    """

    tolerance = var['tolerance']
    y = var['y']

    def calculate_using_residual_measurements_ratio(var):
        """
        Calculate stop criterion based on the residual to measurements ratio.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in calculating of the stop criterion.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        r_norm : float
            The current 2-norm of the residual.

        Notes
        -----
        This stop criterion is based on the ratio of the 2-norm of the current
        residual to the 2-norm of the measurments.

        """

        r_norm = np.linalg.norm(var['r'])
        stop = r_norm < tolerance * np.linalg.norm(y)

        return stop, r_norm

    return calculate_using_residual_measurements_ratio


def get_function_handle(method, var):
    """
    Return a function handle to a given calculation method.

    Parameters
    ----------
    method : str
        Identifier of the calculation method to return a handle to.
    var : dict
        Local variables needed in the calculation method.

    Returns
    -------
    f_handle : function
        Handle to calculation `method` defined in this globals scope.

    """

    return globals()['wrap_calculate_using_' + method](var)



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the `magni.cs.reconstruction.it`
subpackage.

See also
--------
magni.cs.reconstruction._config.Configger : The Configger class used

Notes
-----
This module instantiates the `Configger` class provided by
`magni.utils.config`. The configuration options are the following:

iterations : int
    The maximum number of iterations to do (the default is 300).
kappa : {'fixed', 'adaptive'}
    The method used to calculate the step-size (relaxation) parameter kappa.
kappa_fixed : float
    The step-size (relaxation parameter) used in the algorithm when a fixed
    step-size is used (the default is 0.65).
precision_float : {np.float, np.float16, np.float32, np.float64, np.float128}
    The floating point precision used for the computations (the default is
    np.float64).
report_history : bool
    The indicator of whether or not to return the progress history along with
    the result (the default is False).
stop_criterion : str
    The stop criterion to use in the iterations (the default is
    residual_measurements_ratio).
threshold : {'far', 'fixed'}
    The method used for calculating the threshold value.
threshold_fixed : int
    The number of non-zero coefficients in the signal vector when this number
    is assumed fixed.
threshold_operator : {'hard', 'soft', 'weighted_hard', 'weighted_soft', 'none'}
    The threshold operator used in the backprojection step.
threshold_weights : ndarray
    Array of weights to be used in one of the weighted threshold operators (the
    default is array([[1]]), which implies that all coefficients are weighted
    equally).
tolerance : float
    The least acceptable ratio of residual to measurements (in 2-norm) to break
    the interations (the default is 0.001).
true_solution : ndarray or None
    The true solution to allow for tracking the convergence of the algorithm in
    the artificial setup where the true solution is known a-priori (the default
    is None, which implies that no true solution tracking is used).
warm_start : ndarray
    The initial guess of the solution vector (the default is None, which
    implies that a vector of zeros is used).

"""

from __future__ import division

import numpy as np

from magni.cs.reconstruction._config import Configger as _Configger
from magni.cs.reconstruction.it import _util
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric

configger = _Configger(
    {'iterations': 300,
     'kappa': 'fixed',
     'kappa_fixed': 0.65,
     'precision_float': np.float64,
     'report_history': False,
     'stop_criterion': 'residual_measurements_ratio',
     'threshold': 'far',
     'threshold_fixed': 1,
     'threshold_operator': 'hard',
     'threshold_weights': np.array([[1]]),
     'tolerance': 1e-3,
     'true_solution': None,
     'warm_start': None},
    {'iterations': _numeric(None, 'integer', range_='[1;inf)'),
     'kappa': _generic(
         None, 'string', value_in=_util._get_methods('step_size')),
     'kappa_fixed': _numeric(None, 'floating', range_='(0;inf)'),
     'precision_float': _generic(None, type, value_in=(
         np.float,
         getattr(np, 'float16', np.float_),
         getattr(np, 'float32', np.float_),
         getattr(np, 'float64', np.float_),
         getattr(np, 'float128', np.float_))),
     'report_history': _numeric(None, 'boolean'),
     'stop_criterion': _generic(
         None, 'string', value_in=_util._get_methods('stop_criterion')),
     'threshold': _generic(
         None, 'string', value_in=_util._get_methods('threshold')),
     'threshold_fixed': _numeric(None, 'integer', range_='(0;inf)'),
     'threshold_operator': _generic(
         None, 'string', value_in=_util._get_operators('threshold_operators')),
     'threshold_weights': _numeric(
         None, ('integer', 'floating'), shape=(-1, 1)),
     'tolerance': _numeric(None, 'floating', range_='[0;inf]'),
     'true_solution': _numeric(
         None, ('integer', 'floating', 'complex'), shape=(-1, 1),
         ignore_none=True),
     'warm_start': _numeric(
         None, ('integer', 'floating', 'complex'), shape=(-1, 1),
         ignore_none=True)})



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing output channel functions for the Generalised Approximate
Message Passing (GAMP) algorithm.

Routine listings
----------------
ValidatedMMSEOutputChannel(magni.utils.validation.types.MMSEOutputChannel)
    A base class for validated `magni.cs.reconstruction.gamp` output channels.
AWGN(ValidatedMMSEOutputChannel)
    An Additive White Gaussian Noise (AWGN) MMSE output channel.

"""

from __future__ import division

import numpy as np
import scipy.stats
try:
    import bottleneck as bn
    calculate_median = bn.median
except ImportError:
    calculate_median = np.median

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation import validate_once as _validate_once
from magni.utils.validation.types import (
    MMSEOutputChannel as _MMSEOutputChannel)


class ValidatedMMSEOutputChannel(_MMSEOutputChannel):
    """
    A base class for validated `magni.cs.reconstruction.gamp` output channels.

    Parameters
    ----------
    var : dict
        The output channel state variables.

    """

    def __init__(self, var):
        super(ValidatedMMSEOutputChannel, self).__init__(var)

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()

    @_validate_once
    def compute(self, var):
        """
        Compute the output channel value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the output channel value.

        Returns
        -------
        mean : ndarray
            The computed output channel mean.
        variance : ndarray
            The computed output channel variance.

        Notes
        -----
        This method honors `magni.utils.validation.enable_allow_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()


class AWGN(ValidatedMMSEOutputChannel):
    """
    An Additive White Gaussian Noise (AWGN) MMSE output channel.

    This channel is based on equations (41), (42), and (43) in [1]_ and allows
    for using Expectation Maximization (EM) for learning the channel parameter
    as detailed in equation (77) in [2]_ (see also [2]_ for an introduction to
    EM for GAMP).

    Parameters
    ----------
    sigma_sq : float or int
        The noise level variance (initial noise level when the noise level is
        estimated).
    noise_level_estimation : str
        The method used for estimating (learning) the noise level in each
        iteration.

    Notes
    -----
    The above Parameters are the output channel parameters that must be passed
    in a `var` dict to the channel constructor.

    Possible values for `noise_level_estimation` are:

    * 'sample_variance' - Estimate noise level using the sample variance.
    * 'median' - Estimate noise level from the median.
    * 'em' - Estimate noise using Expectation Maximization (EM).
    * 'fixed' - Use a fixed noise level in all iterations.

    References
    ----------
    .. [1] S. Rangan, "Generalized Approximate Message Passing for Estimation
       with Random Linear Mixing", arXiv:1010.5141v2, pp. 1-22, Aug. 2012.
    .. [2] F. Krzakala, M. Mezard, F. Sausset, Y. Sun, and L. Zdeborova,
       "Probabilistic reconstruction in compressed sensing: algorithms, phase
       diagrams, and threshold achieving matrices", *Journal of Statistical
       Mechanics: Theory and Experiment*, vol. P08009, pp. 1-57, Aug. 2012.
    .. [3] J. P. Vila and P. Schniter, "Expectation-Maximization
       Gaussian-Mixture Approximate Message Passing", *IEEE Transactions on
       Signal Processing*, 2013, vol. 61, no. 19, pp. 4658-4672, Oct. 2013.

    """

    def __init__(self, var):
        super(AWGN, self).__init__(var)

        @_decorate_validation
        def validate_channel_parameters():
            _generic(('var', 'output_channel_parameters'), 'mapping')
            _numeric(('var', 'output_channel_parameters', 'sigma_sq'),
                     ('integer', 'floating'), range_='[0;inf)')
            _generic(('var', 'output_channel_parameters',
                      'noise_level_estimation'),
                     'string',
                     value_in=['sample_variance', 'median', 'em', 'fixed'])

        validate_channel_parameters()

        c_params = var['output_channel_parameters']

        self.m = var['y'].shape[0]
        self.noise_level_estimation = c_params['noise_level_estimation']
        self.stdQ1 = var['convert'](scipy.stats.norm.ppf(1 - 0.25))
        self.sigma_sq = var['convert'](c_params['sigma_sq'])

    def compute(self, var):
        """
        Compute the AWGN output channel value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the output channel value.

        Returns
        -------
        mean : ndarray
            The computed output channel mean.
        variance : ndarray
            The computed output channel variance.

        """

        super(AWGN, self).compute(var)

        y = var['y']

        if var['it'] != 0:  # Allow for fixed initialisation of sigma_sq
            # Noise level estimation for current iteration
            if self.noise_level_estimation == 'sample_variance':
                self.sigma_sq = 1.0/self.m * np.sum(
                    (y - var['A_dot_alpha_bar'])**2)
            elif self.noise_level_estimation == 'median':
                # Estimate variance based on median
                # std(x) ~= median(|x|) / stdQ1 for x~N(0, std**2)
                self.sigma_sq = (
                    calculate_median(np.abs(y - var['A_dot_alpha_bar'])) /
                    self.stdQ1) ** 2

        sigma_sq_plus_v = self.sigma_sq + var['v']

        mean = (var['v'] * y + self.sigma_sq * var['o']) / sigma_sq_plus_v
        variance = (self.sigma_sq * var['v']) / sigma_sq_plus_v

        if self.noise_level_estimation == 'em':
            # EM noise level recursion
            # See Eq. (77) in [1]
            self.sigma_sq = self.sigma_sq * np.sum(
                (np.abs(mean - var['o']) / var['v'])**2) / np.sum(
                    1.0 / sigma_sq_plus_v)

        if len(variance.shape) != 2:
            # When using a sum approximation of A_asq, the variance becomes
            # scalar. However, our interface dictates that it must be a column
            # vector for the sum approximation to be correct. Also, the
            # denominator sum of the sigma_sq EM update must be scaled
            # correctly as if sigma_sq_plus_v had been a vector.

            variance = np.ones(mean.shape, dtype=var['convert']) * variance
            self.sigma_sq = self.sigma_sq / len(y)

        return mean, variance



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the actual simulation functionality.

Routine listings
----------------
run(algorithm, path, label)
     Simulate a reconstruction algorithm.

See Also
--------
magni.cs.phase_transition._config : Configuration options.

Notes
-----
The results of the simulation are backed up throughout the simulation. In case
the simulation is interrupted during execution, the simulation will resume from
the last backup point when run again.

"""

from __future__ import division
import os
import random
import sys
import time

import numpy as np

from magni.cs.phase_transition import _backup
from magni.cs.phase_transition import config as _conf
from magni.cs.phase_transition import _data
from magni.utils.multiprocessing import File as _File
from magni.utils.multiprocessing import process as _process
from magni.utils import split_path as _split_path

if sys.version_info[0] == 2:
    iter_range = xrange(2**30)
else:
    iter_range = range(2**32)


def run(algorithm, path, label, pre_simulation_hook=None):
    """
    Simulate a reconstruction algorithm.

    The simulation results are stored in a HDF5 database rather than returned
    by the function.

    Parameters
    ----------
    algorithm : function
        A function handle to the reconstruction algorithm.
    path : str
        The path of the HDF5 database where the results should be stored.
    label : str
        The label assigned to the simulation results
    pre_simumlation_hook : callable
        A handle to a callable which should be run *just* before the call to
        the reconstruction algorithm (the default is None, which implies that
        no pre hook is run).

    """

    tmp_dir = _split_path(path)[0] + '.tmp' + os.sep
    tmp_file = tmp_dir + label.replace('/', '#') + '.hdf5'

    if not os.path.isdir(tmp_dir):
        os.mkdir(tmp_dir)

    if not os.path.isfile(tmp_file):
        _backup.create(tmp_file)

    done = _backup.get(tmp_file)
    shape = [len(_conf['delta']), len(_conf['rho'])]
    random.seed(_conf['seed'])
    seeds = np.array(
        random.sample(iter_range, shape[0] * shape[1])).reshape(shape)

    tasks = [(algorithm, (i, j), seeds[i, j], tmp_file, pre_simulation_hook)
             for i in range(shape[0]) for j in range(shape[1])
             if not done[i, j]]

    _process(_simulate, args_list=tasks, maxtasks=_conf['maxpoints'])

    with _File(tmp_file, 'r') as f:
        stat_time = f.root.time[:]
        stat_dist = f.root.dist[:]
        stat_mse = f.root.mse[:]
        stat_norm = f.root.norm[:]

    with _File(path, 'a') as f:
        f.create_array('/' + label, 'time', stat_time, createparents=True)
        f.create_array('/' + label, 'dist', stat_dist, createparents=True)
        f.create_array('/' + label, 'mse', stat_mse, createparents=True)
        f.create_array('/' + label, 'norm', stat_norm, createparents=True)

    os.remove(tmp_file)

    if len(os.listdir(tmp_dir)) == 0:
        os.removedirs(tmp_dir)


def _simulate(algorithm, ij_tuple, seed, path, pre_simulation_hook=None):
    """
    Run a number of monte carlo simulations in a single delta-rho point.

    The result of a simulation is the simulation error distance, i.e., the
    ratio between the energy of the coefficient residual and the energy of the
    coefficient vector. The time of the simulation is the execution time of the
    reconstruction attempt.

    Parameters
    ----------
    algorithm : function
        A function handle to the reconstruction algorithm.
    ij_tuple : tuple
        A tuple (i, j) containing the parameters i, j as listed below.
    i : int
        The delta-index of the point in the delta-rho grid.
    j : int
        The rho-index of the point in the delta-rho grid.
    seed : int
        The seed to use in the random number generator when generating the
        problem suite instances.
    path : str
        The path of the HDF5 backup database.
    pre_simulation_hook : callable
        A handle to a callable which should be run *just* before the call to
        the reconstruction algorithm (the default is None, which implies that
        no pre hook is run).

    See Also
    --------
    magni.cs.phase_transition._config: Configuration options.
    magni.cs.phase_transition._data.generate_matrix : Matrix generation.
    magni.cs.phase_transition._data.generate_vector : Coefficient vector
        generation.

    Notes
    -----
    The `pre_simulation_hook` may be used to setup the simulation to match the
    specfic simulation parameters, e.g. if an oracle estimator is used in the
    reconstruction algorithm. The `pre_simulation_hook` takes one argument
    which is the locals() dict.

    The following reconstruction statistics are computed:

    * time: Measured algorithm run time in seconds.
    * dist: Normalised mean squared error (NMSE) - (
      ||alpha_hat - alpha|| / ||alpha||)^2
    * mse: Mean squared error (MSE) - 1/n * ||alpha_hat - alpha||^2
    * norm: True vector norm - ||alpha||

    """

    i, j = ij_tuple

    n = _conf['problem_size']
    m = int(np.round(n * _conf['delta'][i]))
    k = int(np.round(m * _conf['rho'][j]))
    noise = _conf['noise']

    stat_time = np.zeros(_conf['monte_carlo'], dtype=np.float64)
    stat_dist = stat_time.copy()
    stat_mse = stat_time.copy()
    stat_norm = stat_time.copy()

    if k > 0:
        np.random.seed(seed)
        for l in range(_conf['monte_carlo']):
            A = _data.generate_matrix(m, n)
            alpha = _data.generate_vector(n, k)

            if noise is not None:
                e = _data.generate_noise(m, n, k)
                y = A.dot(alpha) + e
            else:
                y = A.dot(alpha)

            if pre_simulation_hook is not None:
                pre_simulation_hook(locals())

            start = time.time()
            alpha_hat = algorithm(y, A, **_conf['algorithm_kwargs'])
            end = time.time()

            error_norm = np.linalg.norm(alpha_hat - alpha)
            alpha_norm = np.linalg.norm(alpha)

            stat_time[l] = end - start
            stat_dist[l] = (error_norm / alpha_norm)**2
            stat_mse[l] = 1.0/n * error_norm**2
            stat_norm[l] = alpha_norm

    _backup.set(path, (i, j), stat_time, stat_dist, stat_mse, stat_norm)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests for `magni.imaging.dictionaries`.

**Testing Strategy**

This test set-up primarily serves to verify the overcomplete transform
capabilities implemented in the above module. The following operations are
tested:

* Whether a vector transformed both back and forth ends up identical to the
  original vector
* A known vector transformed to the secondary domain is identical to the known
  correct result

These cases are tested with 2D transforms.

"""

from __future__ import division
import unittest

import numpy as np
from pkg_resources import parse_version as _parse_version
from scipy import __version__ as _scipy_version

from magni.imaging.dictionaries import get_DCT
from magni.imaging.dictionaries import get_DFT
from magni.imaging import mat2vec
from magni.imaging import vec2mat


_scipy_pre_016 = _parse_version(_scipy_version) < _parse_version('0.16.0')


class TransformsMixin(object):
    """
    Test of overcomplete transforms on a square matrix.

    """

    @unittest.skipIf(_scipy_pre_016, "Not supported for SciPy <= 0.16.0")
    def test_wrong_size_dct(self):
        # Check that the function does not allow specifying an
        # _under_-complete transform
        with self.assertRaises(ValueError):
            matrix = get_DCT(self.array_shape, (self.array_shape[0] - 1,
                                                self.array_shape[1] - 1))

    def test_wrong_size_dft(self):
        # Check that the function does not allow specifying an
        # _under_-complete transform
        with self.assertRaises(ValueError):
            matrix = get_DFT(self.array_shape, (self.array_shape[0] - 1,
                                                self.array_shape[1] - 1))

    @unittest.skipIf(_scipy_pre_016, "Not supported for SciPy <= 0.16.0")
    def test_roundtrip_dct(self):
        # Test DCT
        dct_mtx = get_DCT(self.array_shape, self.array_shape_oc)
        d2_array_dct = dct_mtx.T.dot(mat2vec(self.d2_array))
        d2_array_roundtrip = vec2mat(dct_mtx.dot(d2_array_dct),
                                     self.array_shape)

        # Does the DCT array have the expected shape?
        self.assertSequenceEqual(d2_array_dct.shape,
                                 (self.array_shape_oc[0] *
                                  self.array_shape_oc[1], 1))
        # Is the result identical to the original?
        self.assertTrue(np.allclose(self.d2_array,
                                    d2_array_roundtrip))

    def test_roundtrip_dft(self):
        # Test DFT
        dft_mtx = get_DFT(self.array_shape,
                          self.array_shape_oc)
        d2_array_dft = dft_mtx.conj().T.dot(mat2vec(self.d2_array))
        d2_array_roundtrip = vec2mat(dft_mtx.dot(d2_array_dft),
                                     self.array_shape)

        # Does the DFT array have the expected shape?
        self.assertSequenceEqual(d2_array_dft.shape,
                                 (self.array_shape_oc[0] *
                                  self.array_shape_oc[1], 1))
        # Is the result identical to the original?
        self.assertTrue(np.allclose(self.d2_array,
                                    d2_array_roundtrip))

    @unittest.skipIf(_scipy_pre_016, "Not supported for SciPy <= 0.16.0")
    def test_forward_dct(self):
        # Manually build the DCT-II transform matrices
        # according to the documentation for scipy.fftpack.dct

        idx_freq_0 = np.arange(self.array_shape_oc[0]).reshape((-1, 1))
        idx_freq_1 = np.arange(self.array_shape_oc[1]).reshape((-1, 1))
        idx_space_0 = np.arange(self.array_shape_oc[0]).reshape((1, -1))
        idx_space_1 = np.arange(self.array_shape_oc[1]).reshape((1, -1))

        coeff_mtx_0 = np.pi * idx_freq_0.dot(
            (2 * idx_space_0 + 1)/(2*self.array_shape_oc[0]))
        dct_mtx_0 = 2 * np.cos(coeff_mtx_0)
        # Due to norm='ortho'
        dct_mtx_0[0, :] = (np.sqrt(1 / (4 * self.array_shape_oc[0])) *
                           dct_mtx_0[0, :])
        dct_mtx_0[1:, :] = (np.sqrt(1 / (2 * self.array_shape_oc[0])) *
                            dct_mtx_0[1:, :])

        coeff_mtx_1 = np.pi * idx_freq_1.dot(
            (2 * idx_space_1 + 1) / (2 * self.array_shape_oc[1]))
        dct_mtx_1 = 2 * np.cos(coeff_mtx_1)
        # Due to norm='ortho'
        dct_mtx_1[0, :] = (np.sqrt(1 / (4 * self.array_shape_oc[1])) *
                           dct_mtx_1[0, :])
        dct_mtx_1[1:, :] = (np.sqrt(1 / (2 * self.array_shape_oc[1])) *
                            dct_mtx_1[1:, :])

        # Compute the reference DCT transform
        reference = dct_mtx_0.dot(np.pad(self.d2_array,
                                         ((0, self.array_shape_oc[0] -
                                           self.array_shape[0]),
                                          (0, 0)), 'constant')).T
        reference = dct_mtx_1.dot(np.pad(reference,
                                         ((0, self.array_shape_oc[1] -
                                           self.array_shape[1]),
                                          (0, 0)),  'constant')).T

        # Compute the DCT transform by the tested function
        dct_mtx = get_DCT(self.array_shape, self.array_shape_oc)
        d2_array_dct = vec2mat(dct_mtx.T.dot(mat2vec(self.d2_array)),
                               self.array_shape_oc)

        # Is the result identical to the reference?
        self.assertTrue(np.allclose(reference, d2_array_dct))

    def test_forward_dft(self):
        # Manually build the DFT transform matrices
        # according to https://en.wikipedia.org/wiki/Discrete_Fourier_transform

        idx_freq_0 = np.arange(self.array_shape_oc[0]).reshape((-1, 1))
        idx_freq_1 = np.arange(self.array_shape_oc[1]).reshape((-1, 1))
        idx_space_0 = np.arange(self.array_shape_oc[0]).reshape((1, -1))
        idx_space_1 = np.arange(self.array_shape_oc[1]).reshape((1, -1))

        coeff_mtx_0 = -2 * np.pi * 1j * idx_freq_0.dot(idx_space_0 /
                                                       self.array_shape_oc[0])
        dft_mtx_0 = np.sqrt(1 / self.array_shape_oc[0]) * np.exp(coeff_mtx_0)

        coeff_mtx_1 = -2 * np.pi * 1j * idx_freq_1.dot(idx_space_1 /
                                                       self.array_shape_oc[1])
        dft_mtx_1 = np.sqrt(1 / self.array_shape_oc[1]) * np.exp(coeff_mtx_1)

        # Compute the reference DFT transform
        reference = dft_mtx_0.dot(np.pad(self.d2_array,
                                         ((0, self.array_shape_oc[0] -
                                           self.array_shape[0]),
                                          (0, 0)), 'constant')).T
        reference = dft_mtx_1.dot(np.pad(reference,
                                         ((0, self.array_shape_oc[1] -
                                           self.array_shape[1]),
                                          (0, 0)),  'constant')).T

        # Compute the DFT transform by the tested function
        dft_mtx = get_DFT(self.array_shape, self.array_shape_oc)
        d2_array_dft = vec2mat(dft_mtx.conj().T.dot(mat2vec(self.d2_array)),
                               self.array_shape_oc)

        # Is the result identical to the reference?
        self.assertTrue(np.allclose(reference, d2_array_dft))


class TestTransformSquare(unittest.TestCase, TransformsMixin):
    """
    Test of overcomplete transforms on a square matrix.

    """

    def setUp(self):
        n = 13

        self.d2_array = np.arange(n**2).reshape((n, n))
        self.array_shape = self.d2_array.shape
        self.array_shape_oc = (2*n, 2*n)


class TestTransformNonSquare(unittest.TestCase, TransformsMixin):
    """
    Test of overcomplete transforms on a non-square transform.

    """

    def setUp(self):
        m, n = 17, 23

        self.d2_array = np.arange(m*n).reshape((m, n))
        self.array_shape = self.d2_array.shape
        self.array_shape_oc = (2*m, 2*n)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for visualising dictionary coefficients.

Routine listings
----------------
visualise_DCT(shape)
    Function for visualising DCT coefficients.
visualise_DFT(shape)
    Function for visualising DFT coefficients.

"""

from __future__ import division

import numpy as np

from magni.imaging import vec2mat as _vec2mat


def visualise_DCT(shape):
    """
    Return utilities for visualising DCT coefficients.

    A handle to a function to transform the coefficients into a 'displayable'
    format is returned along with a tuple of ranges of the axes in the 2D
    coefficient plane.

    Parameters
    ----------
    shape : tuple
        The shape of the 2D DCT being visualised.

    Returns
    -------
    display_coefficients : Function
        The function used to transform coefficients into a 'displayable'
        format.
    axes_extent : tuple
        The ranges of the axes in the 2D coefficient plane.

    Notes
    -----
    The display_coefficients function takes log10 to the absolute value of the
    transform cofficient vector given to it as an argument. The returned
    displayable coefficients is a matrix.

    The axes_extent consists of (abcissa_min, abcissa_max, ordinate_min,
    ordinate_max).

    """

    h, w = shape

    def display_coefficients(x):
        return np.log10(np.abs(_vec2mat(x, (h, w))))

    axes_extent = (0, w - 1, h - 1, 0)

    return display_coefficients, axes_extent


def visualise_DFT(shape):
    """
    Return utilities for visualising DFT coefficients.

    A handle to a function to transform the coefficients into a 'displayable'
    format is returned along with a tuple of ranges of the axes in the 2D
    coefficient plane.

    Parameters
    ----------
    shape : tuple
        The shape of the 2D DFT being visualised.

    Returns
    -------
    display_coefficients : Function
        The function used to transform coefficients into a 'displayable'
        format.
    axes_extent : tuple
        The ranges of the axes in the 2D coefficient plane.

    Notes
    -----
    The display_coefficients function takes log10 to the absolute value of the
    transform cofficient vector given to it as an argument. The returned
    displayable coefficients is a matrix that is flipped up/down and
    fftshifted.

    The axes_extent consists of (abcissa_min, abcissa_max, ordinate_min,
    ordinate_max).

    """

    h, w = shape

    def display_coefficients(x):
        return np.flipud(
            np.fft.fftshift(np.log10(np.abs(_vec2mat(x, (h, w))))))

    axes_extent = (-(w // 2), (w - 1) // 2, -(h // 2), (h - 1) // 2)

    return display_coefficients, axes_extent



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing plotting for the `magni.cs.phase_transition` subpackage.

Routine listings
----------------
plot_phase_transitions(curves, plot_l1=True, output_path=None)
    Function for plotting phase transition boundary curves.
plot_phase_transition_colormap(dist, delta, rho, plot_l1=True,
    output_path=None)
    Function for plotting reconstruction probabilities in the phase space.
plot_phase_transition_computation_times(time, delta, rho, output_path=None)
    Function for plotting average computation times in the phase space.

"""

from __future__ import division
from itertools import cycle

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from pkg_resources import parse_version as _parse_version

from magni.cs.phase_transition import config as _conf
from magni.utils.plotting import linestyles as _linestyles
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric

if _parse_version(mpl.__version__) >= _parse_version('1.5.0'):
    _mpl_prop_era = True
else:
    _mpl_prop_era = False


def plot_phase_transitions(curves, plot_l1=True, output_path=None,
                           legend_loc='upper left', errorevery=None,
                           reference_curves=None):
    r"""
    Plot of a set of phase transition boundary curves.

    The set of phase transition boundary curves are plotted an saved under the
    `output_path`, if specified. The `curves` must be a list of dictionaries
    each having keys *delta*, *rho*, and *label*. *delta* must be an `ndarray`
    of :math:`\delta` values in the phase space. *rho* must be an `ndarray` of
    the corresponding :math:`\rho` values in the phase space. *label* must be a
    `str` describing the curve. Optionally, a `curves` dictionary may have an
    *yerr* key with a corresponding 2-by-"len(delta)" array as value. The first
    row in this array specifices the location of the upper error bars whereas
    the second row specifies the location of the lower error bars.

    Parameters
    ----------
    curves : list
        A list of dicts describing the curves to plot.
    plot_l1 : bool, optional
        Whether or not to plot the theoretical :math:`\ell_1` curve (the
        default is True).
    output_path : str, optional
        Path (including file type extension) under which the plot is saved (the
        default value is None, which implies that the plot is not saved).
    legend_loc : str
        Location of legend as a `matplotlib` legend location string (the
        default is 'upper left', which implies that the legend is placed in the
        upper left corner of the plot.)
    errorevery : int
        The subsampling of error bars if used (the default value is None, which
        implies that the default Matplotlib value for errorevery is used.)
    reference_curves : list, optional
        The list of dicts describing the reference curves to plot.

    Notes
    -----
    The plotting is done using `matplotlib`, which implies that an open figure
    containing the plot will result from using this function.

    Tabulated values of the theoretical :math:`\ell_1` phase transition
    boundary is available at
    http://people.maths.ox.ac.uk/tanner/polytopes.shtml

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.cs.phase_transition.plotting import plot_phase_transitions
    >>> delta = np.array([0.1, 0.2, 0.9])
    >>> rho = np.array([0.1, 0.3, 0.8])
    >>> curves = [{'delta': delta, 'rho': rho, 'label': 'data1'}]
    >>> output_path = 'phase_transitions.pdf'
    >>> plot_phase_transitions(curves, output_path=output_path)

    """

    @_decorate_validation
    def validate_input():
        _levels('curves', (_generic(None, 'collection'),
                           _generic(None, 'mapping',
                                    has_keys=('delta', 'rho', 'label'))))

        for i, curve in enumerate(curves):
            _numeric(('curves', i, 'delta'), 'floating', range_='[0;1]',
                     shape=(-1,))
            _numeric(('curves', i, 'rho'), 'floating', range_='[0;1]',
                     shape=(curve['delta'].shape[0],))
            _generic(('curves', i, 'label'), 'string')

            if 'yerr' in curve:
                _numeric(('curves', i, 'yerr'), 'floating', range_='[0;1]',
                         shape=(2, curve['delta'].shape[0]))

        _numeric('plot_l1', 'boolean')
        _generic('output_path', 'string', ignore_none=True)
        _generic('legend_loc', 'string')
        _numeric('errorevery', 'integer', range_='[1;inf)', ignore_none=True)
        _levels('reference_curves', (_generic(None, 'collection',
                                              ignore_none=True),
                                     _generic(None, 'mapping',
                                              has_keys=(
                                                  'delta', 'rho', 'label'))))
        if reference_curves is not None:
            for i, ref_curve in enumerate(reference_curves):
                _numeric(('reference_curves', i, 'delta'), 'floating',
                         range_='[0;1]', shape=(-1,))
                _numeric(('reference_curves', i, 'rho'), 'floating',
                         range_='[0;1]', shape=(ref_curve['delta'].shape[0],))
                _generic(('reference_curves', i, 'label'), 'string')

                if 'style' in ref_curve:
                    _generic(('reference_curves', i, 'style'), 'mapping')

    validate_input()

    fig, axes = plt.subplots(1, 1)
    if not _mpl_prop_era:
        # Emulate matplotlib cycler
        colors = plt.rcParams['axes.color_cycle']
        style_cycle = cycle([{'color': color, 'ls': linestyle}
                             for linestyle in _linestyles for color in colors])
    else:
        # Handled by matplotlib cycler
        style_cycle = cycle([{}])

    error_style = {'capthick': 2}
    if errorevery is not None:
        error_style['errorevery'] = errorevery

    for curve in curves:
        if 'yerr' in curve:
            yerr = [curve['yerr'][0, :] - curve['rho'],
                    curve['rho'] - curve['yerr'][1, :]]
        else:
            yerr = None

        kwargs = dict(next(style_cycle), **error_style)
        axes.errorbar(curve['delta'], curve['rho'], label=curve['label'],
                      yerr=yerr, **kwargs)

    _plot_extra_curves(axes, plot_l1, reference_curves)

    handles, labels = axes.get_legend_handles_labels()
    handles, labels = zip(*sorted(zip(handles, labels),
                                  key=lambda t: t[1].lower()))
    leg = axes.legend(handles, labels, loc=legend_loc)
    leg.get_frame().set_facecolor('1.0')
    axes.set_xlim([0, 1])
    axes.set_ylim([0, 1])
    axes.set_xlabel(r'$\delta = m/n$')
    axes.set_ylabel(r'$\rho = k/m$')

    plt.tight_layout()

    if output_path is not None:
        plt.savefig(output_path)


def plot_phase_transition_colormap(dist, delta, rho, plot_l1=True,
                                   output_path=None):
    r"""
    Create a colormap of the phase space reconstruction probabilities.

    The `delta` and `rho` values span a 2D grid in the phase space.
    Reconstruction probabilities are then calculated from the `dist` 3D array
    of reconstruction error distances. The resulting 2D grid of reconstruction
    probabilites is visualised over the square centers in this 2D grid using a
    colormap. Values in this grid at lower indicies correspond to lower values
    of :math:`\delta` and :math:`\rho`. If `plot_l1` is True, then the
    theoretical l1 curve is overlayed the colormap. The colormap is saved under
    the `output_path`, if specified.

    Parameters
    ----------
    dist : ndarray
        A 3D array of reconstruction error distances.
    delta : ndarray
        :math:`\delta` values used in the 2D grid.
    rho : ndarray
        :math:`\rho` values used in the 2D grid.
    plot_l1 : bool
        Whether or not to plot the theoretical :math:`\ell_1` curve. (the
        default is True)
    output_path : str, optional
        Path (including file type extension) under which the plot is saved (the
        default value is None which implies, that the plot is not saved).

    See Also
    --------
    magni.cs.phase_transition.plotting.plot_phase_transition_computation_times
        Plot average computation times.
    magni.cs.phase_transition.io.load_phase_transition : Loading phase
        transitions from an HDF database.

    Notes
    -----
    The plotting is done using `matplotlib`, which implies that an open figure
    containing the plot will result from using this function.

    The values in `delta` and `rho` are assumed to be equally spaced.

    Due to the *centering* of the color coded rectangles, they are not
    necessarily square towards the ends of the intervals defined by `delta` and
    `rho`.

    Tabulated values of the theoretical :math:`\ell_1` phase transition
    boundary is available at
    http://people.maths.ox.ac.uk/tanner/polytopes.shtml

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.cs.phase_transition.plotting import (
    ...     plot_phase_transition_colormap)
    >>> delta = np.array([0.2, 0.5, 0.8])
    >>> rho = np.array([0.3, 0.6])
    >>> dist = np.array([[[1.35e-08, 1.80e-08], [1.08, 1.11]],
    ... [[1.40e-12, 8.32e-12], [8.57e-01, 7.28e-01]], [[1.92e-13, 1.17e-13],
    ... [2.10e-10,   1.12e-10]]])
    >>> out_path = 'phase_transition_cmap.pdf'
    >>> plot_phase_transition_colormap(dist, delta, rho, output_path=out_path)

    """

    @_decorate_validation
    def validate_input():
        _numeric('delta', 'floating', range_='[0;1]', shape=(-1,))
        _numeric('rho', 'floating', range_='[0;1]', shape=(-1,))
        _numeric('dist', 'floating', range_='[0;inf]',
                 shape=(delta.shape[0], rho.shape[0], -1))
        _numeric('plot_l1', 'boolean')
        _generic('output_path', 'string', ignore_none=True)

    validate_input()

    NMSE_tolerance = 10**(-_conf['SNR'] / 10)
    probs = np.sum(
        dist < NMSE_tolerance, axis=-1, dtype=float) / np.size(dist, -1)

    d_delta = (delta[1] - delta[0]) / 2
    d_rho = (rho[1] - rho[0]) / 2

    x = np.hstack((np.array([0]), delta[1:] - d_delta, np.array([1])))
    y = np.hstack((np.array([0]), rho[1:] - d_rho, np.array([1])))

    X, Y = np.meshgrid(x, y)

    fig, axes = plt.subplots(1, 1)
    p_mesh = axes.pcolormesh(X, Y, probs.T, vmin=0, vmax=1, edgecolor='face')
    c_bar = plt.colorbar(p_mesh)
    c_bar.set_label('Estimated probability of reconstruction')

    # Apply vector graphics render workaround to avoid white gaps in colorbar
    # See: http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.colorbar
    c_bar.solids.set_edgecolor("face")

    if plot_l1:
        _plot_theoretical_l1(axes)
        leg = axes.legend(loc='upper left')
        leg.get_frame().set_facecolor('1.0')

    axes.set_xticks(np.arange(11) / 10)
    axes.set_yticks(np.arange(11) / 10)
    axes.set_xlabel(r'$\delta = m/n$')
    axes.set_ylabel(r'$\rho = k/m$')

    plt.tight_layout()

    if output_path is not None:
        plt.savefig(output_path)


def plot_phase_transition_computation_times(time, delta, rho,
                                            output_path=None):
    r"""
    Create a colormap of the phase space average algorithm computation time.

    The `delta` and `rho` values span a 2D grid in the phase space.
    Average computation times are then calculated from the `time` 3D array of
    reconstruction times. The resulting 2D grid of average reconstruction times
    is visualised over the square centers in this 2D grid using a colormap.
    Values in this grid at lower indicies correspond to lower values of
    :math:`\delta` and :math:`\rho`. The colormap is saved under the
    `output_path`, if specified.

    Parameters
    ----------
    time : ndarray
        A 3D array of algorithm computation times.
    delta : ndarray
        :math:`\delta` values used in the 2D grid.
    rho : ndarray
        :math:`\rho` values used in the 2D grid.
    output_path : str, optional
        Path (including file type extension) under which the plot is saved (the
        default value is None which implies, that the plot is not saved).

    See Also
    --------
    magni.cs.phase_transition.plotting.plot_phase_transition_colormap : Plot
        phase transition probabilities.
    magni.cs.phase_transition.io.load_phase_transition : Loading phase
        transitions from an HDF database.

    Notes
    -----
    The plotting is done using `matplotlib`, which implies that an open figure
    containing the plot will result from using this function.

    The values in `delta` and `rho` are assumed to be equally spaced.

    Due to the *centering* of the color coded rectangles, they are not
    necessarily square towards the ends of the intervals defined by `delta` and
    `rho`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.cs.phase_transition.plotting import (
    ...     plot_phase_transition_computation_times)
    >>> delta = np.array([0.2, 0.5, 0.8])
    >>> rho = np.array([0.3, 0.6])
    >>> times = np.array([[[1.35e-08, 1.80e-08], [1.08, 1.11]],
    ... [[1.40e-12, 8.32e-12], [8.57e-01, 7.28e-01]], [[1.92e-13, 1.17e-13],
    ... [2.10e-10,   1.12e-10]]])
    >>> out_path = 'computation_times_cmap.pdf'
    >>> plot_phase_transition_computation_times(
    ...     times, delta, rho, output_path=out_path)

    """

    @_decorate_validation
    def validate_input():
        _numeric('delta', 'floating', range_='[0;1]', shape=(-1,))
        _numeric('rho', 'floating', range_='[0;1]', shape=(-1,))
        _numeric('time', 'floating', range_='[0;inf]',
                 shape=(delta.shape[0], rho.shape[0], -1))
        _generic('output_path', 'string', ignore_none=True)

    validate_input()

    mean_times = time.mean(axis=-1)

    d_delta = (delta[1] - delta[0]) / 2
    d_rho = (rho[1] - rho[0]) / 2

    x = np.hstack((np.array([0]), delta[1:] - d_delta, np.array([1])))
    y = np.hstack((np.array([0]), rho[1:] - d_rho, np.array([1])))

    X, Y = np.meshgrid(x, y)

    fig, axes = plt.subplots(1, 1)
    p_mesh = axes.pcolormesh(X, Y, mean_times.T, edgecolor='face')
    c_bar = plt.colorbar(p_mesh)
    c_bar.set_label('Average computation time [s]')

    # Apply vector graphics render workaround to avoid white gaps in colorbar
    # See: http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.colorbar
    c_bar.solids.set_edgecolor("face")

    axes.set_xticks(np.arange(11) / 10)
    axes.set_yticks(np.arange(11) / 10)
    axes.set_xlabel(r'$\delta = m/n$')
    axes.set_ylabel(r'$\rho = k/m$')

    plt.tight_layout()

    if output_path is not None:
        plt.savefig(output_path)


def _plot_extra_curves(axes, plot_l1, reference_curves):
    """
    Plot any extra curves that might be needed in a phase space plot.

    Parameters
    ----------
    axes : matplotlib.axes.Axes
        The matplotlib Axes instance on which the theoretical l1 phase
        transition should be plotted.
    plot_l1 : bool, optional
        Whether or not to plot the theoretical :math:`\ell_1` curve (the
        default is True).
    reference_curves : list
        The list of dicts describing the reference curves to plot.

    """

    if plot_l1:
        # Add theoretical ell_1 plot
        _plot_theoretical_l1(axes)

    if reference_curves is not None:
        # Add additional reference curves
        for ref_curve in reference_curves:
            axes.plot(ref_curve['delta'], ref_curve['rho'],
                      label=ref_curve['label'], **ref_curve['style'])


def _plot_theoretical_l1(axes):
    """
    Plot the theoretical l1 phase transition on the `axes`.

    Parameters
    ----------
    axes : matplotlib.axes.Axes
        The matplotlib Axes instance on which the theoretical l1 phase
        transition should be plotted.

    Notes
    -----
    The plotted theoretical :math:`\ell1` phase transition is based on
    tabulated values of available at
    http://people.maths.ox.ac.uk/tanner/polytopes.shtml

    """

    delta = [0, 0, 0, 0.001, 0.008, 0.021, 0.038, 0.058, 0.078, 0.1, 0.122,
             0.144, 0.167, 0.19, 0.212, 0.235, 0.258, 0.282, 0.305, 0.329,
             0.352, 0.375, 0.399, 0.422, 0.445, 0.468, 0.49, 0.513, 0.535,
             0.558, 0.58, 0.603, 0.626, 0.647, 0.669, 0.691, 0.712, 0.733,
             0.754, 0.774, 0.794, 0.813, 0.832, 0.851, 0.868, 0.884, 0.9,
             0.915, 0.929, 0.942, 0.953, 0.963, 0.972, 0.98, 0.986, 0.991,
             0.996, 0.998, 1]
    rho = [0, 0.025, 0.051, 0.077, 0.103, 0.125, 0.144, 0.16, 0.176, 0.19,
           0.202, 0.215, 0.227, 0.238, 0.249, 0.26, 0.271, 0.282, 0.293, 0.304,
           0.315, 0.326, 0.337, 0.348, 0.359, 0.37, 0.381, 0.392, 0.404, 0.415,
           0.428, 0.44, 0.453, 0.466, 0.479, 0.493, 0.507, 0.522, 0.537, 0.552,
           0.568, 0.585, 0.602, 0.621, 0.639, 0.658, 0.678, 0.699, 0.721,
           0.744, 0.767, 0.791, 0.815, 0.84, 0.864, 0.891, 0.921, 0.949, 1]

    axes.plot(delta, rho, c='k', ls='--', label=r'Theoretical $\ell_1$')



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
lissajous_sample_image(h, w, scan_length, num_points, f_y=1., f_x=1.,
    theta_y=0., theta_x=np.pi / 2)
    Function for lissajous sampling an image.
lissajous_sample_surface(l, w, speed, sample_rate, time, f_y=1., f_x=1.,
    theta_y=0., theta_x=np.pi / 2, speed_mode=0)
    Function for lissajous sampling a surface.

"""

from __future__ import division

import numpy as np

from magni.imaging.measurements import _util
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['lissajous_sample_image', 'lissajous_sample_surface']

_min_l = _util.min_l
_min_w = _util.min_w
_min_speed = _util.min_speed
_min_sample_rate = _util.min_sample_rate
_min_time = _util.min_time
_min_scan_length = _util.min_scan_length
_min_num_points = _util.min_num_points


def lissajous_sample_image(h, w, scan_length, num_points, f_y=1., f_x=1.,
                           theta_y=0., theta_x=np.pi / 2):
    """
    Sample an image using a lissajous pattern.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a lissajous pattern are determined. The
    `scan_length` determines the length of the path scanned whereas
    `num_points` indicates the number of samples taken on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.
    f_y : float
        The frequency of the y-sinusoid (the default value is 1.0).
    f_x : float
        The frequency of the x-sinusoid (the default value is 1.0).
    theta_y : float
        The starting phase of the y-sinusoid (the default is 0.0).
    theta_x : float
        The starting phase of the x-sinusoid (the default is pi / 2).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import lissajous_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> np.set_printoptions(suppress=True)
    >>> lissajous_sample_image(h, w, scan_length, num_points)
    array([[ 5.        ,  9.5       ],
           [ 1.40370042,  7.70492686],
           [ 0.67656563,  3.75183526],
           [ 3.39871123,  0.79454232],
           [ 7.39838148,  1.19240676],
           [ 9.48459832,  4.62800824],
           [ 7.99295651,  8.36038857],
           [ 4.11350322,  9.41181634],
           [ 0.94130617,  6.94345168],
           [ 1.0071768 ,  2.92458128],
           [ 4.25856283,  0.56150128],
           [ 8.10147506,  1.7395012 ],
           [ 9.4699986 ,  5.51876059]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')
        _numeric('scan_length', 'floating',
                 range_='[{};inf)'.format(_min_scan_length))
        _numeric('num_points', 'integer',
                 range_='[{};inf)'.format(_min_num_points))
        _numeric('f_y', 'floating', range_='(0;inf)')
        _numeric('f_x', 'floating', range_='(0;inf)')
        _numeric('theta_y', 'floating', range_='(-inf;inf)')
        _numeric('theta_x', 'floating', range_='(-inf;inf)')

    validate_input()

    coords = lissajous_sample_surface(
        float(h - 1), float(w - 1), scan_length, float(num_points), 1.,
        f_y=f_y, f_x=f_x, theta_y=theta_y, theta_x=theta_x)
    coords = coords + 0.5

    return coords


def lissajous_sample_surface(l, w, speed, sample_rate, time, f_y=1., f_x=1.,
                             theta_y=0., theta_x=np.pi / 2, speed_mode=0):
    """
    Sample a surface area using a lissajous pattern.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using a lissajous pattern are determined. The scanned
    path is determined from the probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.
    f_y : float
        The frequency of the y-sinusoid (the default value is 1.0).
    f_x : float
        The frequency of the x-sinusoid (the default value is 1.0).
    theta_y : float
        The starting phase of the y-sinusoid (the default is 0.0).
    theta_x : float
        The starting phase of the x-sinusoid (the default is pi / 2).
    speed_mode : int
        The speed mode used to select sampling points (the default is 0 which
        implies that the speed argument determines the speed, and f_y and f_x
        determine the ratio between the relative frequencies used).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis.

    Generally, the lissajous sampling pattern does not provide constant speed,
    and this cannot be compensated for without violating f_y, f_x, or both.
    Therefore, `speed_mode` allows the user to determine how this issue is
    handled: In `speed_mode` 0, constant speed equal to `speed` is ensured by
    non-uniform sampling of a lissajous curve, whereby `f_y` and `f_x` are not
    constant frequencies. In `speed_mode` 1, average speed equal to `speed` is
    ensured by scaling `f_y` and `f_x` by the same constant. In `speed_mode` 2,
    `f_y` and `f_x` are kept constant and the `speed` is only used to determine
    the path length in combination with `time`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import lissajous_sample_surface
    >>> l = 1e-6
    >>> w = 1e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> np.set_printoptions(suppress=True)
    >>> lissajous_sample_surface(l, w, speed, sample_rate, time)
    array([[ 0.0000005 ,  0.000001  ],
           [ 0.00000001,  0.00000058],
           [ 0.00000033,  0.00000003],
           [ 0.00000094,  0.00000025],
           [ 0.00000082,  0.00000089],
           [ 0.00000017,  0.00000088],
           [ 0.00000007,  0.00000024],
           [ 0.00000068,  0.00000003],
           [ 0.00000099,  0.0000006 ],
           [ 0.00000048,  0.000001  ],
           [ 0.        ,  0.00000057],
           [ 0.00000035,  0.00000002],
           [ 0.00000094,  0.00000027]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('l', 'floating', range_='[{};inf)'.format(_min_l))
        _numeric('w', 'floating', range_='[{};inf)'.format(_min_w))
        _numeric('speed', 'floating', range_='[{};inf)'.format(_min_speed))
        _numeric('sample_rate', 'floating',
                 range_='[{};inf)'.format(_min_sample_rate))
        _numeric('time', 'floating', range_='[{};inf)'.format(_min_time))
        _numeric('f_y', 'floating', range_='(0;inf)')
        _numeric('f_x', 'floating', range_='(0;inf)')
        _numeric('theta_y', 'floating', range_='(-inf;inf)')
        _numeric('theta_x', 'floating', range_='(-inf;inf)')
        _numeric('speed_mode', 'integer', range_='[0;2]')

    validate_input()

    s_x = w / 2
    s_y = l / 2

    if speed_mode in (0, 1):
        # The probe moves 4 * s_x * f_x and 4 * s_y * f_y pixels a second in
        # the x-direction and y-direction, respectively, and the 2-norm of this
        # is a lower bound on the distance per second. Thus, t is an upper
        # bound on the scan time.
        t = speed * time / np.sqrt((4 * s_x * f_x)**2 + (4 * s_y * f_y)**2)
        # The above assumes that f_x * t and f_y * t are integral numbers and
        # so t is increased to ensure the upper bound.
        t = max(np.ceil(f_x * t) / f_x, np.ceil(f_y * t) / f_y)
        # The distance between sampling points on the curve is chosen small
        # enough to approximate the curve by straight line segments.
        dt = 1 / (10**4 * max(f_x, f_y))
        t = np.linspace(0, t, int(t / dt))

        x = s_x * np.cos(2 * np.pi * f_x * t + theta_x) + s_x
        y = s_y * np.cos(2 * np.pi * f_y * t + theta_y) + s_y
        dx = x[1:] - x[:-1]
        dy = y[1:] - y[:-1]
        l = np.zeros(t.shape)
        l[1:] = np.cumsum((dx**2 + dy**2)**(1 / 2))

        if speed_mode == 0:
            # Constant speed entails constant distance between samples.
            l_mode_0 = np.linspace(0, speed * time, sample_rate * time + 1)
            t = np.interp(l_mode_0, l, t)
        else:  # speed_mode == 1
            # The value of t where the desired scan length is reached.
            t_end = np.argmax(l > speed * time) * dt
            t = np.linspace(0, t_end, sample_rate * time + 1)
    else:  # speed_mode == 2
        t = np.linspace(0, time, sample_rate * time + 1)

    x = s_x * np.cos(2 * np.pi * f_x * t + theta_x) + s_x
    y = s_y * np.cos(2 * np.pi * f_y * t + theta_y) + s_y

    return np.column_stack((x, y))



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for calculating the starting value of L (the number
of gradient descent iterations for each sigma) used in the Smoothed l0
algorithms.

Routine listings
----------------
calculate_using_fixed(var)
    Calculate the fixed L value.
calculate_using_geometric(var)
    Calculate an L value in a 'geometric' way.
get_function_handle(method)
    Return a function handle to a given calculation method.

"""

from __future__ import division


def wrap_calculate_using_fixed(var):
    """
    Arguments wrapper for calculate_using_fixed.

    """

    convert = var['convert']
    L = convert(var['param']['L_fixed'])

    def calculate_using_fixed():
        """
        Calculate the fixed L value.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the L value.

        Returns
        -------
        L : float
            The L value to be used in the SL0 algorithm.

        """

        return L

    return calculate_using_fixed


def wrap_calculate_using_geometric(var):
    """
    Arguments wrapper for calculate_using_geometric.

    """

    convert = var['convert']
    L_start = convert(var['param']['L_geometric_start'])

    def calculate_using_geometric():
        """
        Calculate an L value in a 'geometric' way.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the L value.

        Returns
        -------
        L : float
            The L value to be used in the SL0 algorithm.

        """

        return L_start

    return calculate_using_geometric


def get_function_handle(method, var):
    """
    Return a function handle to a given calculation method.

    Parameters
    ----------
    method : str
        Identifier of the calculation method to return a handle to.
    var : dict
        Local variables needed in the L method.

    Returns
    -------
    f_handle : function
        Handle to the calculation method defined in this globals scope.

    """

    return globals()['wrap_calculate_using_' + method](var)



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing implementations of Approximate Message Passing (AMP).

Routine listings
----------------
config
    Configger providing configuration options for this subpackage.
threshold_operator
    Module providing AMP threshold operators.
run(y, A)
    Run the AMP reconstruction algorithm.
stop_criterion
    Module providing AMP stop criteria.
util
    Module providing AMP utilities.


Notes
-----
An implementation of the Donoho, Maleki, Montanari (DMM) AMP using soft
thresholding from [1]_ is provided with threshold options as specified in [2]_.

The implementation allows for using custom thresholding functions and stop
criteria.

References
----------
.. [1] D. L. Donoho, A. Maleki, and A. Montanari, "Message-passing algorithms
   for compressed sensing", *Proceedings of the National Academy of Sciences of
   the United States of America*, vol. 106, no. 45, p. 18914-18919, Nov. 2009.
.. [2] A. Montanari, "Graphical models concepts in compressed sensing" *in
   Compressed Sensing: Theory and Applications*, Y. C. Eldar and G. Kutyniok
   (Ed.), Cambridge University Press, ch. 9, pp. 394-438, 2012.

"""

from magni.cs.reconstruction.amp import (threshold_operator, stop_criterion)
from magni.cs.reconstruction.amp._config import configger as config
from magni.cs.reconstruction.amp._algorithm import run
import magni.cs.reconstruction.amp.util



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing data container classes for .mi spectroscopy files.

The classes of this module can be used either directly or indirectly through
the `io` module by loading an .mi spectroscopy file.

Routine listings
----------------
Buffer(magni.afmtypes.BaseClass)
    Data class for .mi spectroscopy buffer.
Chunk(magni.afmtypes.BaseClass)
    Data class for .mi spectroscopy chunk.
Grid(magni.afmtypes.BaseClass)
    Data class for .mi spectroscopy grid.
Point(magni.afmtypes.BaseClass)
    Data class for .mi spectroscopy point.
Spectroscopy(magni.afm.types.File)
    Data class for .mi spectroscopy.

See Also
--------
magni.afm.io : .mi file loading.

"""

from __future__ import division
from collections import OrderedDict as _OrderedDict

import numpy as np

from magni.afm.types import BaseClass as _BaseClass
from magni.afm.types import File as _File
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


class Buffer(_BaseClass):
    """
    Data class of the .mi spectroscopy buffers.

    Parameters
    ----------
    attrs : dict
        The attributes of the buffer.
    data : list or tuple
        The grids, points, or chunks of the buffer.

    Attributes
    ----------
    data : numpy.ndarray
        The grids, points, or chunks of the buffer.

    See Also
    --------
    magni.utils.types.BaseClass : Superclass of the present class.

    Examples
    --------
    No example .mi spectroscopy file is distributed with magni.

    """

    _params = _OrderedDict((('bufferLabel', str),
                            ('bufferUnit', str)))

    def __init__(self, attrs, data):
        @_decorate_validation
        def validate_input():
            _generic('attrs', 'mapping', has_keys=('bufferLabel',))
            _levels('data', (_generic(None, 'explicit collection'),
                             _generic(None, (Grid, Point, Chunk))))

        _BaseClass.__init__(self, attrs)
        validate_input()

        self._data = tuple(data)

    data = property(lambda self: self._data)


class Chunk(_BaseClass):
    """
    Data class of the .mi spectroscopy chunks.

    Parameters
    ----------
    attrs : dict
        The attributes of the chunk.
    data : numpy.ndarray
        The data of the chunk.

    Attributes
    ----------
    data : numpy.ndarray
        The data of the chunk.

    See Also
    --------
    magni.utils.types.BaseClass : Superclass of the present class.

    Examples
    --------
    No example .mi spectroscopy file is distributed with magni.

    """

    _params = _OrderedDict((('index', int),
                            ('samples', int),
                            ('timeStart', float),
                            ('timeDelta', float),
                            ('sweepStart', float),
                            ('sweepDelta', float),
                            ('continuation', bool),
                            ('label', str),
                            ('pointIndex', int),
                            ('curveIndex', int)))

    def __init__(self, attrs, data):
        @_decorate_validation
        def validate_input():
            _generic('attrs', 'mapping',
                     has_keys=('samples', 'timeStart', 'timeDelta',
                               'sweepStart', 'sweepDelta'))
            _numeric('data', 'floating', shape=(self.attrs['samples'],))

        _BaseClass.__init__(self, attrs)
        validate_input()

        self._time = None
        self._sweep = None
        self._data = data
        self._data.flags['WRITEABLE'] = False

    data = property(lambda self: self._data)

    @property
    def sweep(self):
        """
        Get the sweep property of the chunk.

        The sweep property is the series of the entity which was swept.

        Returns
        -------
        sweep : numpy.ndarray
            The sweep property.

        Notes
        -----
        To reduce the memory footprint of chunks, the series does not exist
        explicitly, until it is requested.

        """

        if self._sweep is None:
            self._sweep = np.linspace(
                self.attrs['sweepStart'],
                self.attrs['sweepStart'] +
                (self.attrs['samples'] - 1) * self.attrs['sweepDelta'],
                self.attrs['samples'])
            self._sweep.flags['WRITEABLE'] = False

        return self._sweep

    @property
    def time(self):
        """
        Get the time property of the chunk.

        Returns
        -------
        time : numpy.ndarray
            The time property.

        Notes
        -----
        To reduce the memory footprint of chunks, the series does not exist
        explicitly, until it is requested.

        """

        if self._time is None:
            self._time = np.linspace(
                self.attrs['timeStart'],
                self.attrs['timeStart'] +
                (self.attrs['samples'] - 1) * self.attrs['timeDelta'],
                self.attrs['samples'])
            self._time.flags['WRITEABLE'] = False

        return self._time


class Grid(_BaseClass):
    """
    Data class of the .mi spectroscopy grids.

    Parameters
    ----------
    attrs : dict
        The attributes of the grid.
    points : list or tuple
        The points of the grid.

    Attributes
    ----------
    points : tuple
        The points of the grid.

    See Also
    --------
    magni.utils.types.BaseClass : Superclass of the present class.

    Notes
    -----
    The points are input and output as a 2D tuple of Point instances.

    Examples
    --------
    No example .mi spectroscopy file is distributed with magni.

    """

    _params = _OrderedDict((('index', int),
                            ('xCenter', float),
                            ('yCenter', float),
                            ('xPoints', int),
                            ('yPoints', int),
                            ('pointSpacing', float),
                            ('angle', int),
                            ('yDirection', str),
                            ('xDirection', str)))

    def __init__(self, attrs, points):
        @_decorate_validation
        def validate_input():
            _generic('attrs', 'mapping', has_keys=(
                'xCenter', 'yCenter', 'xPoints', 'yPoints', 'pointSpacing'))
            _levels('points', (_generic(None, 'explicit collection',
                                        len_=self.attrs['yPoints']),
                               _generic(None, 'explicit collection',
                                        len_=self.attrs['xPoints']),
                               _generic(None, Point, ignore_none=True)))

        _BaseClass.__init__(self, attrs)
        validate_input()

        self._points = tuple(tuple(val) for val in points)

    points = property(lambda self: self._points)


class Point(_BaseClass):
    """
    Data class of the .mi spectroscopy points.

    Parameters
    ----------
    attrs : dict
        The attributes of the point.
    chunks : list or tuple, optional
        The chunks of the point. (the default is (), which implies no chunks)

    Attributes
    ----------
    chunks : tuple
        The chunks of the point.

    See Also
    --------
    magni.utils.types.BaseClass : Superclass of the present class.

    Examples
    --------
    No example .mi spectroscopy file is distributed with magni.

    """

    _params = _OrderedDict((('index', int),
                            ('xCoordinate', float),
                            ('yCoordinate', float)))

    def __init__(self, attrs, chunks=()):
        @_decorate_validation
        def validate_input():
            _generic('attrs', 'mapping', has_keys=tuple(Point._params.keys()))
            _levels('chunks', (_generic(None, 'explicit collection'),
                               _generic(None, Chunk)))

        _BaseClass.__init__(self, attrs)
        validate_input()

        self._chunks = tuple(chunks)

    chunks = property(lambda self: self._chunks)


class Spectroscopy(_File):
    """
    Data class of the .mi spectroscopy files.

    Parameters
    ----------
    attrs : dict
        The attributes of the image.
    buffers : list or tuple
        The buffers of the image.

    See Also
    --------
    magni.utils.types.File : Superclass of the present class.

    Examples
    --------
    No example .mi spectroscopy file is distributed with magni.

    """

    _params = _OrderedDict((('plotType', str),
                            ('BgImageFile', str)))

    def __init__(self, attrs, buffers):
        @_decorate_validation
        def validate_input():
            _levels('buffers', (_generic(None, 'explicit collection'),
                                _generic(None, Buffer)))

        _File.__init__(self, attrs, buffers)
        validate_input()



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing implementations of generic reconstruction algorithms.

Each subpackage provides a family of generic reconstruction algorithms. Thus
each subpackage has a config module and a run function which provide the
interface of the given family of reconstruction algorithms.

Routine listings
----------------
amp
    Subpackage providing implementations of Approximate Message Passing (AMP).
gamp
    Subpackage providing implementations of Generalised Approximate Message
    Passing (GAMP).
it
    Subpackage providing implementations of Iterative Thresholding (IT).
iht
    Subpackage providing implementations of Iterative Hard Thresholding (IHT).
    (Deprecated)
sl0
    Subpackage providing implementations of Smoothed l0 Norm (SL0).

"""

from magni.cs.reconstruction import amp
from magni.cs.reconstruction import gamp
from magni.cs.reconstruction import it
from magni.cs.reconstruction import iht
from magni.cs.reconstruction import sl0



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing threshold functions for the Approximate Message Passing (AMP)
algorithm.

Routine listings
----------------
ValidatedThresholdOperator(magni.utils.validation.types.ThresholdOperator)
    A base class for validated `magni.cs.reconstruction.amp` threshold operator
SoftThreshold(ValidatedThresholdOperator)
    A soft threshold operator.

"""

from __future__ import division

import numpy as np
import scipy.stats
try:
    import bottleneck as bn
    calculate_median = bn.median
except ImportError:
    calculate_median = np.median

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation import validate_once as _validate_once
from magni.utils.validation.types import (
    ThresholdOperator as _ThresholdOperator)


class ValidatedThresholdOperator(_ThresholdOperator):
    """
    A base class for validated `magni.cs.reconstruction.amp` threshold operator

    Parameters
    ----------
    var : dict
        The threshold operator state variables.

    """

    def __init__(self, var):
        super(ValidatedThresholdOperator, self).__init__(var)

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()

    @_validate_once
    def compute_deriv_threshold(self, var):
        """
        Compute the entrywise derivative threshold.

        Parameters
        ----------
        var : dict
            The variables used in computing the derivative threshold.

        Returns
        -------
        eta_deriv : ndarray
            The computed entrywise derivative threshold.

        Notes
        -----
        This method honors `magni.utils.validation.enable_allow_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()

    @_validate_once
    def compute_threshold(self, var):
        """
        Compute the entrywise threshold.

        Parameters
        ----------
        var : dict
            The variables used in computing the threshold.

        Returns
        -------
        eta : ndarray
            The computed entrywise threshold.

        Notes
        -----
        This method honors `magni.utils.validation.enable_allow_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()

    @_validate_once
    def update_threshold_level(self, var):
        """
        Update the threshold level state.

        Parameters
        ----------
        var : dict
            The variables used in computing the threshold level update.

        Notes
        -----
        This method honors `magni.utils.validation.enable_allow_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()


class SoftThreshold(ValidatedThresholdOperator):
    """
    A soft threshold operator.

    This soft threshold operator is based on the description of it and its use
    in AMP as given in [1]_ with corrections from [2]_.

    Parameters
    ----------
    threshold_level_update_method : {'residual', 'median'}
        The method to use for updating the threshold level.
    theta : float
        The tunable regularisation parameter in the threshold level.
    tau_hat_sq : float
        The mean squared error of the (approximated) un-thresholded
        estimate used to determine the threshold level.

    Notes
    -----
    The above Parameters are the threshold parameters that must be passed
    in a `var` dict to the threshold constructor.

    References
    ----------
    .. [1] A. Montanari, "Graphical models concepts in compressed sensing" *in
       Compressed Sensing: Theory and Applications*, Y. C. Eldar and
       G. Kutyniok (Ed.), Cambridge University Press, ch. 9, pp. 394-438, 2012.
    .. [2] J. T. Parker, "Approximate Message Passing Algorithms for
       Generalized Bilinear Inference", PhD Thesis, Graduate School of The Ohio
       State University, 2014

    """

    def __init__(self, var):
        super(SoftThreshold, self).__init__(var)

        @_decorate_validation
        def validate_threshold_parameters():
            _generic(('var', 'threshold_parameters'), 'mapping')
            _generic(('var', 'threshold_parameters',
                      'threshold_level_update_method'), 'string',
                     value_in=['residual', 'median'])
            _numeric(('var', 'threshold_parameters', 'theta'),
                     ('integer', 'floating'), range_='(0;inf)')
            _numeric(('var', 'threshold_parameters', 'tau_hat_sq'),
                     ('integer', 'floating'), range_='[0;inf)')

        validate_threshold_parameters()

        t_params = var['threshold_parameters']

        self.theta = t_params['theta']  # alpha in Eq. (9.44) in [1].
        self.tau_hat_sq = var['convert'](t_params['tau_hat_sq'])
        self.update_method = t_params['threshold_level_update_method']
        self.m = var['y'].shape[0]
        self.stdQ1 = var['convert'](scipy.stats.norm.ppf(1 - 0.25))

    def compute_deriv_threshold(self, var):
        """
        Compute the entrywise derivative soft threshold.

        Parameters
        ----------
        var : dict
            The variables used in computing the derivative threshold.

        Returns
        -------
        eta_deriv : ndarray
            The computed entrywise derivative threshold.

        """

        super(SoftThreshold, self).compute_deriv_threshold(var)

        op = var['alpha_bar_prev'] + var['AH_dot_chi']
        thres = self.theta * np.sqrt(self.tau_hat_sq)
        eta_deriv = var['convert']((op > thres) + (op < -thres))

        return eta_deriv

    def compute_threshold(self, var):
        """
        Compute the entrywise soft threshold.

        Parameters
        ----------
        var : dict
            The variables used in computing the threshold.

        Returns
        -------
        eta : ndarray
            The computed entrywise threshold.

        """

        super(SoftThreshold, self).compute_threshold(var)

        op = var['alpha_bar'] + var['AH_dot_chi']
        thres = self.theta * np.sqrt(self.tau_hat_sq)
        eta = var['convert'](
            ((op - thres) * (op > thres) + (op + thres) * (op < -thres)))

        return eta

    def update_threshold_level(self, var):
        """
        Update the threshold level state.

        Parameters
        ----------
        var : dict
            The variables used in computing the threshold level update.

        """

        super(SoftThreshold, self).update_threshold_level(var)

        chi = var['chi']

        if self.update_method == 'residual':
            # Eq. (9.44) in [1]
            self.tau_hat_sq = 1.0 / self.m * np.linalg.norm(chi)**2

        elif self.update_method == 'median':
            # Eq. (9.45) in [1] corrected according to [2]
            self.tau_hat_sq = (
                1.0 / self.stdQ1 * calculate_median(np.abs(chi)))**2
